# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"Base class that implements the base builder flow"

import abc
import logging
import os
import os.path as p
from threading import Lock
from typing import Any, Dict, FrozenSet, Iterable, Optional, Set, Tuple

from hdl_checker.database import Database  # pylint: disable=unused-import
from hdl_checker.diagnostics import CheckerDiagnostic, DiagType
from hdl_checker.exceptions import SanityCheckError
from hdl_checker.parsers.elements.identifier import Identifier
from hdl_checker.path import Path
from hdl_checker.types import (
    BuildFlags,
    BuildFlagScope,
    FileType,
    RebuildInfo,
    RebuildLibraryUnit,
    RebuildPath,
    RebuildUnit,
)


class BaseBuilder(object):  # pylint: disable=useless-object-inheritance
    """
    Class that implements the base builder flow
    """

    __metaclass__ = abc.ABCMeta

    # Set an empty container for the default flags
    default_flags = {
        BuildFlagScope.dependencies: {},
        BuildFlagScope.single: {},
        BuildFlagScope.all: {},
    }  # type: Dict[BuildFlagScope, Dict[FileType, BuildFlags]]

    _external_libraries = {FileType.vhdl: set(), FileType.verilog: set()}  # type: dict

    _include_paths = {FileType.vhdl: set(), FileType.verilog: set()}  # type: dict

    @classmethod
    def addExternalLibrary(cls, lang, library_name):
        # type: (FileType, Identifier) -> None
        """
        Adds an external library so it may be referenced by the builder
        directly
        """
        assert lang in cls._external_libraries, "Uknown language '%s'" % lang.value
        cls._external_libraries[lang].add(library_name)

    @classmethod
    def addIncludePath(cls, lang, path):
        # type: (FileType, str) -> None
        """
        Adds an include path to be used by the builder where applicable
        """
        assert lang in cls._include_paths, "Uknown language '%s'" % lang.value
        cls._include_paths[lang].add(path)

    @abc.abstractproperty
    def builder_name(self):
        # type: (...) -> Any
        """
        Defines the builder identification
        """

    @abc.abstractproperty
    def file_types(self):
        # type: (...) -> Any
        """
        Returns the file types supported by the builder
        """

    def __init__(self, work_folder, database):
        # type: (Path, Database) -> None
        # Shell accesses must be atomic
        self._lock = Lock()

        self._logger = logging.getLogger(__package__ + "." + self.builder_name)
        self._database = database
        self._work_folder = p.abspath(p.expanduser(work_folder.name))
        self._build_info_cache = {}  # type: Dict[Path, Dict[str, Any]]
        self._builtin_libraries = None  # type: Optional[Set[Identifier]]
        self._added_libraries = set()  # type: Set[Identifier]

        self.setup()

    def setup(self):
        # type: (...) -> Any
        """
        Creates directories and parses builtins libraries
        """
        # Fallback builder has no setup
        if self.builder_name == "fallback":
            return

        if not p.exists(self._work_folder):
            self._logger.debug("Target folder '%s' was created", self._work_folder)
            os.makedirs(self._work_folder)
        else:
            self._logger.debug("%s already exists", self._work_folder)

        self.checkEnvironment()

    @classmethod
    def __jsonDecode__(cls, state):
        # type: (...) -> Any
        """
        Returns an object of cls based on a given state
        """
        # pylint: disable=protected-access
        obj = super(BaseBuilder, cls).__new__(cls)
        obj._logger = logging.getLogger(state.pop("_logger"))
        obj._builtin_libraries = set(state.pop("_builtin_libraries"))
        obj._added_libraries = set(state.pop("_added_libraries"))

        obj._lock = Lock()
        obj._build_info_cache = {}
        obj.__dict__.update(state)
        # pylint: enable=protected-access

        return obj

    def __jsonEncode__(self):
        # type: (...) -> Any
        """
        Gets a dict that describes the current state of this object
        """
        state = self.__dict__.copy()
        state["_logger"] = self._logger.name
        state["_builtin_libraries"] = list(self.builtin_libraries)
        state["_added_libraries"] = list(self._added_libraries)
        del state["_build_info_cache"]
        del state["_lock"]
        return state

    @staticmethod
    def isAvailable():  # pragma: no cover
        # type: (...) -> Any
        """
        Method that should be overriden by child classes and return True
        if the given builder is available on the current environment
        """
        raise NotImplementedError

    def checkEnvironment(self):
        # type: (...) -> Any
        """
        Sanity environment check for child classes. Any exception raised
        is translated to SanityCheckError exception.
        """
        try:
            self._checkEnvironment()
        except Exception as exc:
            raise SanityCheckError(self.builder_name, str(exc))

    @abc.abstractmethod
    def _shouldIgnoreLine(self, line):
        """
        Method called for each stdout output and should return True if
        the given line should not be parsed using _makeRecords
        and _searchForRebuilds
        """

    @abc.abstractmethod
    def _makeRecords(self, line):
        """
        Static method that converts a string into a dict that has
        elements identifying its fields
        """

    def _getRebuilds(self, path, line, library):
        # type: (Path, str, Identifier) -> Set[RebuildInfo]
        """
        Gets info on what should be rebuilt to satisfy the builder
        """
        try:
            parse_results = self._searchForRebuilds(line)
        except NotImplementedError:  # pragma: no cover
            return set()

        rebuilds = set()  # type: Set[RebuildInfo]
        for rebuild in parse_results:
            unit_type = rebuild.get("unit_type", None)
            library_name = rebuild.get("library_name", None)
            unit_name = rebuild.get("unit_name", None)
            rebuild_path = rebuild.get("rebuild_path", None)

            if None not in (unit_type, unit_name):
                for dependency in self._database.getDependenciesByPath(path):
                    if dependency.name.name == rebuild["unit_name"]:
                        rebuilds.add(RebuildUnit(unit_name, unit_type))
                        break
            elif None not in (library_name, unit_name):
                if library_name == "work":
                    library_name = library.name
                rebuilds.add(RebuildLibraryUnit(unit_name, library_name))
            elif rebuild_path is not None:
                # GHDL sometimes gives the full path of the file that
                # should be recompiled
                rebuilds.add(RebuildPath(rebuild_path))
            else:  # pragma: no cover
                self._logger.warning("Don't know what to do with %s", rebuild)

        return rebuilds

    def _searchForRebuilds(self, line):  # pragma: no cover
        # type: (...) -> Any
        """
        Finds units that the builders is telling us to rebuild
        """
        raise NotImplementedError

    def _parseBuiltinLibraries(self):
        # type: (...) -> Iterable[Identifier]
        """
        Discovers libraries that exist regardless before we do anything
        """
        raise NotImplementedError

    @property
    def work_folder(self):
        # type: (...) -> str
        """
        Returns the path to the work folder this builder is using
        """
        return self._work_folder

    @property
    def builtin_libraries(self):
        # type: (...) -> FrozenSet[Identifier]
        """
        Return a list of precompiled libraries this builder is aware of
        """
        if self._builtin_libraries is None:
            try:
                self._builtin_libraries = set(self._parseBuiltinLibraries())
                if self._logger.isEnabledFor(logging.DEBUG):  # pragma: no cover
                    if self._builtin_libraries:  # pragma: no cover
                        self._logger.debug(
                            "Builtin libraries: %s", tuple(self._builtin_libraries)
                        )
                    else:  # pragma: no cover
                        self._logger.debug("No builtin libraries found")
            except NotImplementedError:
                self._builtin_libraries = set()

        return frozenset(self._builtin_libraries)

    @abc.abstractmethod
    def _checkEnvironment(self):
        """
        Sanity environment check that should be implemented by child
        classes. Nothing is done with the return, the child class should
        raise an exception by itself
        """

    @abc.abstractmethod
    def _buildSource(self, path, library, flags=None):
        # type: (Path, Identifier, Optional[BuildFlags]) -> Iterable[str]
        """
        Callback called to actually build the source
        """

    def _getFlags(self, path, scope):
        # type: (Path, BuildFlagScope) -> BuildFlags
        """
        Gets flags to build the path, both builder based and from the database.
        If a build is forced, assume we're building a single file (not its
        dependencies)
        """
        return tuple(self._database.getFlags(path, scope)) + tuple(
            self.default_flags.get(scope, {}).get(FileType.fromPath(path), ())
            + self.default_flags.get(BuildFlagScope.all, {}).get(
                FileType.fromPath(path), ()
            )
        )

    def _buildAndParse(
        self, path, library, flags
    ):  # type: (Path, Identifier, BuildFlags) -> Tuple[Set[CheckerDiagnostic],Set[RebuildInfo]]
        """
        Runs _buildSource method and parses the output to find message
        records and units that should be rebuilt
        """
        if library is None:
            library = self._database.getLibrary(path)  # or Identifier("work")

        self._createLibraryIfNeeded(library)

        for lib in (x.library for x in self._database.getDependenciesByPath(path)):
            self._createLibraryIfNeeded(lib or Identifier("work"))

        diagnostics = set()  # type: Set[CheckerDiagnostic]
        rebuilds = set()  # type: Set[RebuildInfo]

        for line in self._buildSource(path, library, flags=flags):
            if self._shouldIgnoreLine(line):
                continue

            for _record in self._makeRecords(line):
                try:
                    diagnostics |= {_record}
                except:
                    self._logger.exception(
                        " - %s hash: %s | %s",
                        _record,
                        _record.__hash__,
                        type(_record).__mro__,
                    )
                    raise
            rebuilds |= self._getRebuilds(path, line, library)

        # If no filename is set, assume it's for the current path
        for diag in diagnostics:
            if diag.filename is None:
                diag.filename = path

        self._logBuildResults(diagnostics, rebuilds)

        return diagnostics, rebuilds

    def _logBuildResults(self, diagnostics, rebuilds):  # pragma: no cover
        # type: (...) -> Any
        """
        Logs diagnostics and rebuilds only for debugging purposes
        """
        if not self._logger.isEnabledFor(logging.DEBUG):
            return

        if diagnostics:
            self._logger.debug("Diagnostic messages found")
            for record in diagnostics:
                self._logger.debug(record)

        if rebuilds:
            self._logger.debug("Rebuilds found")
            for rebuild in rebuilds:
                self._logger.debug(rebuild)

    def _createLibraryIfNeeded(self, library):
        # type: (Identifier) -> None
        """
        Proxy for only creating libraries once and avoid overwriting builtin
        libraries
        """
        if library in self._added_libraries:
            return
        if library in self.builtin_libraries:
            return
        self._added_libraries.add(library)
        self._createLibrary(library)

    @abc.abstractmethod
    def _createLibrary(self, library):
        # type: (...) -> Any
        """
        Callback called to create a library
        """

    def _isFileTypeSupported(self, path):
        # type: (Path) -> bool
        """
        Checks if a given path is supported by this builder
        """
        return FileType.fromPath(path) in self.file_types

    def build(self, path, library, scope, forced=False):
        # type: (Path, Identifier, BuildFlagScope, bool) -> Tuple[Set[CheckerDiagnostic], Set[RebuildInfo]]
        """
        Method that interfaces with parents and implements the building
        chain
        """

        if not self._isFileTypeSupported(path):
            self._logger.warning(
                "Path '%s' with file type '%s' is not " "supported",
                path,
                FileType.fromPath(path),
            )
            return set(), set()

        if path not in self._build_info_cache:
            self._build_info_cache[path] = {
                "compile_time": 0.0,
                "diagnostics": [],
                "rebuilds": [],
            }

        cached_info = self._build_info_cache[path]

        build = False
        if forced:
            build = True
            self._logger.info("Forcing build of %s", str(path))
        elif path.mtime > cached_info["compile_time"]:
            build = True
            self._logger.info("Building %s", str(path))

        if build:
            with self._lock:
                diagnostics, rebuilds = self._buildAndParse(
                    path, library, self._getFlags(path, scope)
                )

            cached_info["diagnostics"] = diagnostics
            cached_info["rebuilds"] = rebuilds
            cached_info["compile_time"] = path.mtime

            if DiagType.ERROR in [x.severity for x in diagnostics]:
                cached_info["compile_time"] = 0

        else:
            self._logger.debug("Nothing to do for %s", path)
            diagnostics = cached_info["diagnostics"]
            rebuilds = cached_info["rebuilds"]

        return diagnostics, rebuilds



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"Common stuff"

import abc
import functools
import logging
import os
import os.path as p
import shutil
import signal
import subprocess as subp
import sys
from collections import Counter
from tempfile import NamedTemporaryFile
from threading import Lock
from typing import Callable, Dict, Iterable, List, Optional, Tuple, TypeVar, Union

import six

from hdl_checker.path import Path

_logger = logging.getLogger(__name__)

ON_WINDOWS = os.name == "nt"
ON_LINUX = sys.platform == "linux"
ON_MAC = sys.platform == "darwin"


def setupLogging(stream, level, color=True):  # pragma: no cover
    "Setup logging according to the command line parameters"

    if isinstance(stream, six.string_types):

        class Stream(object):  # pylint: disable=useless-object-inheritance
            """
            File subclass that allows RainbowLoggingHandler to write
            with colors
            """

            _lock = Lock()
            _color = color

            def __init__(self, *args, **kwargs):
                self._fd = open(*args, **kwargs)

            def isatty(self):
                """
                Tells if this stream accepts control chars
                """
                return self._color

            def write(self, text):
                """
                Writes to the stream
                """
                with self._lock:
                    self._fd.write(toBytes(text))

        _stream = Stream(stream, "ab", buffering=0)
    else:
        _stream = stream

    try:
        # This is mostly for debugging when doing stuff directly from a
        # terminal
        from rainbow_logging_handler import RainbowLoggingHandler  # type: ignore

        handler = RainbowLoggingHandler(_stream)
    except ImportError:  # pragma: no cover
        handler = logging.StreamHandler(_stream)
        handler.formatter = logging.Formatter(
            "%(levelname)-7s | %(asctime)s | "
            + "%(name)s @ %(funcName)s():%(lineno)d %(threadName)s "
            + "|\t%(message)s",
            datefmt="%H:%M:%S",
        )

    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("pynvim").setLevel(logging.WARNING)
    logging.getLogger("pyls").setLevel(logging.INFO)
    logging.getLogger("pyls.python_ls").setLevel(logging.INFO)
    logging.getLogger("pyls.config.config").setLevel(logging.WARNING)
    logging.getLogger("matplotlib").setLevel(logging.INFO)
    logging.getLogger("pyls_jsonrpc.endpoint").setLevel(logging.INFO)
    logging.root.addHandler(handler)
    logging.root.setLevel(level)


# From here: http://stackoverflow.com/a/8536476/1672783
def terminateProcess(pid):
    "Terminate a process given its PID"

    if ON_WINDOWS:
        import ctypes

        process_terminate = 1
        handle = ctypes.windll.kernel32.OpenProcess(process_terminate, False, pid)
        ctypes.windll.kernel32.TerminateProcess(handle, -1)
        ctypes.windll.kernel32.CloseHandle(handle)
    else:
        os.kill(pid, signal.SIGTERM)


def isProcessRunning(pid):
    "Checks if a process is running given its PID"

    if ON_WINDOWS:
        return _isProcessRunningOnWindows(pid)

    return _isProcessRunningOnPosix(pid)


def _isProcessRunningOnPosix(pid):
    "Checks if a given PID is runnning under POSIX OSs"
    try:
        os.kill(pid, 0)
    except OSError:
        return False

    return True


def _isProcessRunningOnWindows(pid):
    """
    Enumerates active processes as seen under windows Task Manager on Win
    NT/2k/XP using PSAPI.dll (new api for processes) and using ctypes.Use it as
    you please.

    Based on information from
    http://support.microsoft.com/default.aspx?scid=KB;EN-US;Q175030&ID=KB;EN-US;Q175030

    By Eric Koome email ekoome@yahoo.com
    license GPL

    (adapted from code found at
    http://code.activestate.com/recipes/305279-getting-process-information-on-windows/)
    """
    from ctypes import windll, c_ulong, sizeof, byref

    # PSAPI.DLL
    psapi = windll.psapi

    arr = c_ulong * 256
    list_of_pids = arr()
    cb = sizeof(list_of_pids)  # pylint: disable=invalid-name
    cb_needed = c_ulong()

    # Call Enumprocesses to get hold of process id's
    psapi.EnumProcesses(byref(list_of_pids), cb, byref(cb_needed))

    # Number of processes returned
    number_of_pids = int(cb_needed.value / sizeof(c_ulong()))

    pid_list = [i for i in list_of_pids][:number_of_pids]

    return int(pid) in pid_list


if not hasattr(p, "samefile"):

    def _samefile(file1, file2):
        """
        Emulated version of os.path.samefile. This is needed for Python
        2.7 running on Windows (at least on Appveyor CI)
        """

        return os.stat(file1) == os.stat(file2)


else:
    _samefile = p.samefile  # pylint: disable=invalid-name

samefile = _samefile  # pylint: disable=invalid-name


def removeDuplicates(seq):
    """
    Fast removal of duplicates within an iterable
    """
    seen = set()
    seen_add = seen.add

    return [x for x in seq if not (x in seen or seen_add(x))]


# Copied from ycmd
def toBytes(value):  # pragma: no cover
    """
    Consistently returns the new bytes() type from python-future.
    Assumes incoming strings are either UTF-8 or unicode (which is
    converted to UTF-8).
    """

    if not value:
        return bytes()

    # This is tricky. On py2, the bytes type from builtins (from python-future) is
    # a subclass of str. So all of the following are true:
    #   isinstance(str(), bytes)
    #   isinstance(bytes(), str)
    # But they don't behave the same in one important aspect: iterating over a
    # bytes instance yields ints, while iterating over a (raw, py2) str yields
    # chars. We want consistent behavior so we force the use of bytes().

    if isinstance(value, bytes):
        return value

    # This is meant to catch Python 2's native str type.

    if isinstance(value, bytes):
        return bytes(value, encoding="utf8")

    if isinstance(value, str):
        # On py2, with `from builtins import *` imported, the following is true:
        #
        #   bytes(str(u'abc'), 'utf8') == b"b'abc'"
        #
        # Obviously this is a bug in python-future. So we work around it. Also filed
        # upstream at: https://github.com/PythonCharmers/python-future/issues/193
        # We can't just return value.encode('utf8') on both py2 & py3 because on
        # py2 that *sometimes* returns the built-in str type instead of the newbytes
        # type from python-future.

        if six.PY2:
            return bytes(value.encode("utf8"), encoding="utf8")

        return bytes(value, encoding="utf8")

    # This is meant to catch `int` and similar non-string/bytes types.

    return toBytes(str(value))


def getTemporaryFilename(name):
    """
    Gets a temporary filename following the format 'hdl_checker_pid<>.log' on Linux
    and 'hdl_checker_pid<>_<unique>.log' on Windows
    """
    basename = "hdl_checker_" + name + "_pid{}".format(os.getpid())

    if ON_WINDOWS:
        return NamedTemporaryFile(
            prefix=basename + "_", suffix=".log", delete=False
        ).name

    return p.join(p.sep, "tmp", basename + ".log")


def isFileReadable(path):
    # type: (Path) -> bool
    """
    Checks if a given file is readable
    """
    try:
        open(str(path), "r").close()

        return True
    except IOError:
        return False


def runShellCommand(cmd_with_args, shell=False, env=None, cwd=None):
    # type: (Union[Tuple[str], List[str]], bool, Optional[Dict], Optional[str]) -> Iterable[str]
    """
    Runs a shell command and handles stdout catching
    """
    _logger.debug(" ".join(cmd_with_args))

    try:
        stdout = list(
            subp.check_output(
                cmd_with_args,
                stderr=subp.STDOUT,
                shell=shell,
                env=env or os.environ,
                cwd=cwd,
            ).splitlines()
        )
    except subp.CalledProcessError as exc:
        stdout = list(exc.output.splitlines())
        _logger.debug(
            "Command '%s' failed with error code %d.\nStdout:\n%s",
            cmd_with_args,
            exc.returncode,
            "\n".join([x.decode() for x in stdout]),
        )
    except OSError as exc:
        _logger.debug("Command '%s' failed with %s", cmd_with_args, exc)
        raise

    return [x.decode() for x in stdout]


def removeIfExists(filename):
    # type: (str) -> bool
    "Removes filename using os.remove and catches the exception if that fails"
    try:
        os.remove(filename)
        _logger.debug("Removed %s", filename)
        return True
    except OSError:
        _logger.debug("Failed to remove %s", filename)
        return False


def removeDirIfExists(dirname):
    # type: (str) -> bool
    """
    Removes the directory dirname using shutil.rmtree and catches the exception
    if that fails
    """
    try:
        shutil.rmtree(dirname)
        _logger.debug("Removed %s", dirname)
        return True
    except OSError:
        _logger.debug("Failed to remove %s", dirname)
        return False


class HashableByKey(object):  # pylint: disable=useless-object-inheritance
    """
    Implements hash and comparison operators properly across Python 2 and 3
    """

    __metaclass__ = abc.ABCMeta

    @abc.abstractproperty
    def __hash_key__(self):
        """ Implement this attribute to use it for hashing and comparing"""

    def __hash__(self):
        #  return hash(self.__hash_key__)
        try:
            return hash(self.__hash_key__)
        except:
            print("Couldn't hash %s" % repr(self.__hash_key__))
            raise

    def __eq__(self, other):
        """Overrides the default implementation"""

        if isinstance(other, self.__class__):
            return self.__hash_key__ == other.__hash_key__

        return NotImplemented  # pragma: no cover

    def __ne__(self, other):  # pragma: no cover
        """Overrides the default implementation (unnecessary in Python 3)"""
        result = self.__eq__(other)

        if result is not NotImplemented:
            return not result

        return NotImplemented


def logCalls(func):  # pragma: no cover
    # type: (Callable) -> Callable
    "Decorator to Log calls to func"
    import pprint

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        # type: (...) -> Callable
        _str = "%s(%s, %s)" % (func.__name__, args, pprint.pformat(kwargs))
        try:
            result = func(self, *args, **kwargs)
            _logger.info("%s => %s", _str, repr(result))

            return result
        except:
            _logger.exception("Failed to run %s", _str)
            raise

    return wrapper


T = TypeVar("T")  # pylint: disable=invalid-name


def getMostCommonItem(items):
    # type: (Iterable[T]) -> T
    """
    Gets the most common item on an interable of items
    """
    data = Counter(items)
    return max(items, key=data.get)


if six.PY2:

    def readFile(path):
        "Wrapper around open().read() that return \n for new lines"
        return open(path, mode="rU").read()


else:

    def readFile(path):
        "Wrapper around open().read() that return \n for new lines"
        return open(path, mode="r", newline="\n").read()



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=missing-docstring
# pylint: disable=wrong-import-position
# pylint: disable=no-self-use
# pylint: disable=protected-access
# pylint: disable=function-redefined

import logging
import os
import os.path as p
import time
from threading import Event, Timer

import parameterized  # type: ignore
import six
import unittest2  # type: ignore
from mock import MagicMock, patch
import pyls  # type: ignore
from pyls import lsp as defines
from pyls import uris
from pyls.workspace import Workspace  # type: ignore
from pyls_jsonrpc.streams import JsonRpcStreamReader  # type: ignore

from nose2.tools import such  # type: ignore

from hdl_checker.tests import assertCountEqual, getTestTempPath, setupTestSuport  # isort:skip


# Debouncing will hurt testing since it won't actually call the debounced
# function if we call it too quickly.
def _debounce(interval_s, keyed_by=None):  # pylint: disable=unused-argument
    def wrapper(func):
        def debounced(*args, **kwargs):
            result = func(*args, **kwargs)
            _logger.info("%s(%s, %s) returned %s", func.__name__, args, kwargs, result)
            return result

        return debounced

    return wrapper


# Mock debounce before it's applied
pyls._utils.debounce = _debounce

from hdl_checker import lsp  # isort:skip
from hdl_checker.diagnostics import CheckerDiagnostic, DiagType  # isort:skip
from hdl_checker.utils import ON_WINDOWS  # isort:skip

_logger = logging.getLogger(__name__)

JSONRPC_VERSION = "2.0"
LSP_MSG_TEMPLATE = {"jsonrpc": JSONRPC_VERSION, "id": 1, "processId": None}

LSP_MSG_EMPTY_RESPONSE = {"jsonrpc": JSONRPC_VERSION, "id": 1, "result": None}

MOCK_WAIT_TIMEOUT = 5

TEST_TEMP_PATH = getTestTempPath(__name__)
TEST_PROJECT = p.join(TEST_TEMP_PATH, "test_project")

if ON_WINDOWS:
    TEST_PROJECT = TEST_PROJECT.lower()


class TestCheckerDiagToLspDict(unittest2.TestCase):
    @parameterized.parameterized.expand(
        [
            (DiagType.INFO, defines.DiagnosticSeverity.Hint),
            (DiagType.STYLE_INFO, defines.DiagnosticSeverity.Hint),
            (DiagType.STYLE_WARNING, defines.DiagnosticSeverity.Information),
            (DiagType.STYLE_ERROR, defines.DiagnosticSeverity.Information),
            (DiagType.WARNING, defines.DiagnosticSeverity.Warning),
            (DiagType.ERROR, defines.DiagnosticSeverity.Error),
            (DiagType.NONE, defines.DiagnosticSeverity.Error),
        ]
    )
    def test_converting_to_lsp(self, diag_type, severity):
        _logger.info("Running %s and %s", diag_type, severity)

        diag = CheckerDiagnostic(
            checker="hdl_checker test",
            text="some diag",
            filename="filename",
            line_number=1,
            column_number=1,
            error_code="error code",
            severity=diag_type,
        )

        self.assertEqual(
            lsp.checkerDiagToLspDict(diag),
            {
                "source": "hdl_checker test",
                "range": {
                    "start": {"line": 0, "character": 0},
                    "end": {"line": 0, "character": 0},
                },
                "message": "some diag",
                "severity": severity,
                "code": "error code",
            },
        )

    def test_workspace_notify(self):
        workspace = MagicMock(spec=Workspace)

        server = lsp.HdlCodeCheckerServer(workspace, root_path=None)

        server._handleUiInfo("some info")  # pylint: disable=protected-access
        workspace.show_message.assert_called_once_with(
            "some info", defines.MessageType.Info
        )
        workspace.show_message.reset_mock()

        server._handleUiWarning("some warning")  # pylint: disable=protected-access
        workspace.show_message.assert_called_once_with(
            "some warning", defines.MessageType.Warning
        )
        workspace.show_message.reset_mock()

        server._handleUiError("some error")  # pylint: disable=protected-access
        workspace.show_message.assert_called_once_with(
            "some error", defines.MessageType.Error
        )


with such.A("LSP server") as it:
    if six.PY2:
        it.assertCountEqual = assertCountEqual(it)

    def _initializeServer(server, params=None):
        it.assertEqual(
            server.m_initialize(**(params or {})),
            {"capabilities": {"textDocumentSync": 1}},
        )

        it.assertEqual(server.m_initialized(), None)

    def startLspServer():
        _logger.debug("Creating server")
        tx_r, tx_w = os.pipe()
        it.tx = JsonRpcStreamReader(os.fdopen(tx_r, "rb"))

        it.rx = MagicMock()
        it.rx.closed = False

        it.server = lsp.HdlCheckerLanguageServer(it.rx, os.fdopen(tx_w, "wb"))

    def stopLspServer():
        _logger.debug("Shutting down server")
        msg = LSP_MSG_TEMPLATE.copy()
        msg.update({"method": "exit"})
        it.server._endpoint.consume(msg)

        # Close the pipe from the server to stdout and empty any pending
        # messages
        it.tx.close()
        it.tx.listen(_logger.fatal)

        del it.server

    def _waitOnMockCall(meth):
        event = Event()

        timer = Timer(MOCK_WAIT_TIMEOUT, event.set)
        timer.start()

        result = []
        while not event.isSet():
            if meth.mock_calls:
                result = meth.mock_calls[0]
                timer.cancel()
                break

            time.sleep(0.1)

        if event.isSet():
            it.fail("Timeout waiting for %s" % meth)

        return result

    def checkLintFileOnOpen(source):
        return checkLintFileOnMethod(source, "m_text_document__did_open")

    def checkLintFileOnSave(source):
        return checkLintFileOnMethod(source, "m_text_document__did_save")

    def checkLintFileOnMethod(source, method):
        with patch.object(it.server.workspace, "publish_diagnostics"):
            _logger.info("Sending %s request", method)
            getattr(it.server, method)(
                textDocument={"uri": uris.from_fs_path(source), "text": None}
            )

            call = _waitOnMockCall(it.server.workspace.publish_diagnostics)
            doc_uri, diagnostics = call[1]
            _logger.info("doc_uri: %s", doc_uri)
            _logger.info("diagnostics: %s", diagnostics)

            it.assertEqual(doc_uri, uris.from_fs_path(source))
            return diagnostics

    @it.has_setup
    def setup():
        setupTestSuport(TEST_TEMP_PATH)

    with it.having("root URI set but no project file"):

        @it.has_setup
        def setup():
            startLspServer()

        @it.has_teardown
        def teardown():
            stopLspServer()

        import hdl_checker

        @it.should("search for files on initialization")
        @patch.object(hdl_checker.hdl_checker_base.HdlCodeCheckerBase, "configure")
        @patch.object(hdl_checker.config_generators.base_generator.BaseGenerator, "generate")
        def test(configure, generate):
            _initializeServer(
                it.server, params={"rootUri": uris.from_fs_path(TEST_PROJECT)}
            )
            configure.assert_called_once()
            generate.assert_called_once()

        @it.should("lint file when opening it")  # type: ignore
        def test():
            source = p.join(TEST_PROJECT, "another_library", "foo.vhd")

            it.assertCountEqual(
                checkLintFileOnOpen(source),
                [
                    {
                        "source": "HDL Checker/static",
                        "range": {
                            "start": {"line": 28, "character": 11},
                            "end": {"line": 28, "character": 11},
                        },
                        "message": "Signal 'neat_signal' is never used",
                        "severity": defines.DiagnosticSeverity.Information,
                    }
                ],
            )

        @it.should("not lint file outside workspace")  # type: ignore
        def test():
            source = p.join(TEST_PROJECT, "basic_library", "clock_divider.vhd")
            with it.assertRaises(AssertionError):
                checkLintFileOnSave(source)

    with it.having("an existing and valid old style project file"):

        @it.has_setup
        def setup():
            startLspServer()

        @it.has_teardown
        def teardown():
            stopLspServer()

        @it.should("respond capabilities upon initialization")  # type: ignore
        def test():
            _initializeServer(
                it.server,
                params={
                    "rootUri": uris.from_fs_path(TEST_PROJECT),
                    "initializationOptions": {"project_file": "vimhdl.prj"},
                },
            )

        @it.should("lint file when opening it")  # type: ignore
        def test():
            source = p.join(TEST_PROJECT, "basic_library", "clk_en_generator.vhd")

            it.assertCountEqual(checkLintFileOnOpen(source), [])

    with it.having("an existing and valid JSON config file"):

        @it.has_setup
        def setup():
            startLspServer()

        @it.has_teardown
        def teardown():
            stopLspServer()

        @it.should("respond capabilities upon initialization")  # type: ignore
        def test():
            _initializeServer(
                it.server,
                params={
                    "rootUri": uris.from_fs_path(TEST_PROJECT),
                    "initializationOptions": {"project_file": "config.json"},
                },
            )
            #  it.assertTrue(it.server._checker.database._paths)

        @it.should("lint file when opening it")  # type: ignore
        def test():
            source = p.join(TEST_PROJECT, "basic_library", "clk_en_generator.vhd")
            it.assertCountEqual(checkLintFileOnOpen(source), [])

        #  @it.should("clean up if the project file has been modified") # type: ignore
        #  def test():
        #      source = p.join(TEST_PROJECT, "basic_library", "clk_en_generator.vhd")

        #      og_timestamp = p.getmtime(it.server._checker.project_file)

        #      content = open(it.server._checker.project_file).read()
        #      open(it.server._checker.project_file, "a").write("\n# foo")
        #      curr_ts = p.getmtime(it.server._checker.project_file)

        #      _logger.info(
        #          "Original timestamp: %s, current timestamp: %s", og_timestamp, curr_ts
        #      )

        #      it.assertNotEqual(og_timestamp, curr_ts, "Timestamps are still equal??")

        #      with patch.object(it.server.workspace, "publish_diagnostics"):
        #          with patch.object(it.server._checker, "clean") as clean:

        #              _logger.info("Sending m_text_document__did_save request")
        #              it.server.m_text_document__did_save(
        #                  textDocument={"uri": uris.from_fs_path(source), "text": None}
        #              )

        #              call = _waitOnMockCall(it.server.workspace.publish_diagnostics)
        #              doc_uri, diagnostics = call[1]
        #              _logger.info("doc_uri: %s", doc_uri)
        #              _logger.info("diagnostics: %s", diagnostics)

        #              clean.assert_called_once()

        #      # Restore the original content (which will change the timestamp)
        #      # and request a message so that parsing occurs here
        #      open(it.server._checker.project_file, "w").write(content)
        #      it.assertEqual(content, open(it.server._checker.project_file).read())

        #      it.server.m_text_document__did_save(
        #          textDocument={"uri": uris.from_fs_path(source), "text": None}
        #      )

        #  @it.should("rebuild if the cache file has been removed") # type: ignore
        #  def test():
        #      # type: (...) -> Any
        #      server = startLspServer()

        #      source = p.join(TEST_PROJECT, "basic_library", "clk_en_generator.vhd")

        #      os.remove(str(it.server._checker._getCacheFilename()))

        #      def getBuilderName():
        #          return "msim"

        #      with patch.object(it.server.workspace, "publish_diagnostics"):
        #          with patch.object(it.server._checker, "clean") as clean:
        #          #      with patch.object(
        #          #          it.server._checker.config_parser,
        #          #          "getBuilderName",
        #          #          getBuilderName,
        #          #      ):

        #              _logger.info("Sending m_text_document__did_save request")
        #              it.server.m_text_document__did_save(
        #                  textDocument={
        #                      "uri": uris.from_fs_path(source),
        #                      "text": None,
        #                  }
        #              )

        #              call = _waitOnMockCall(it.server.workspace.publish_diagnostics)
        #              doc_uri, diagnostics = call[1]
        #              _logger.info("doc_uri: %s", doc_uri)
        #              _logger.info("diagnostics: %s", diagnostics)

        #              clean.assert_called()

    with it.having("a non existing project file"):

        @it.has_setup
        def setup():
            startLspServer()

        @it.has_teardown
        def teardown():
            stopLspServer()

        @it.should("respond capabilities upon initialization")  # type: ignore
        def test():
            it.project_file = "__some_project_file.prj"
            it.assertFalse(p.exists(it.project_file))

            _initializeServer(
                it.server,
                params={
                    "rootUri": uris.from_fs_path(TEST_PROJECT),
                    "initializationOptions": {"project_file": it.project_file},
                },
            )

        @it.should("lint file when opening it")  # type: ignore
        def test():
            source = p.join(TEST_PROJECT, "another_library", "foo.vhd")
            diagnostics = checkLintFileOnOpen(source)

            it.assertCountEqual(
                diagnostics,
                [
                    {
                        "source": "HDL Checker/static",
                        "range": {
                            "start": {"line": 28, "character": 11},
                            "end": {"line": 28, "character": 11},
                        },
                        "message": "Signal 'neat_signal' is never used",
                        "severity": defines.DiagnosticSeverity.Information,
                    }
                ],
            )

    with it.having("neither root URI nor project file set"):

        @it.has_setup
        def setup():
            startLspServer()

        @it.has_teardown
        def teardown():
            stopLspServer()

        @it.should("respond capabilities upon initialization")  # type: ignore
        def test():
            _initializeServer(
                it.server, params={"initializationOptions": {"project_file": None}}
            )

        @it.should("lint file when opening it")  # type: ignore
        def test():
            source = p.join(TEST_PROJECT, "basic_library", "clock_divider.vhd")

            it.assertCountEqual(
                checkLintFileOnOpen(source),
                [
                    {
                        "source": "HDL Checker/static",
                        "range": {
                            "start": {"line": 26, "character": 11},
                            "end": {"line": 26, "character": 11},
                        },
                        "message": "Signal 'clk_enable_unused' is never used",
                        "severity": defines.DiagnosticSeverity.Information,
                    }
                ],
            )

    with it.having("no root URI but project file set"):

        @it.has_setup
        def setup():
            startLspServer()

        @it.has_teardown
        def teardown():
            stopLspServer()

        @it.should("respond capabilities upon initialization")  # type: ignore
        def test():
            # In this case, project file is an absolute path, since there's no
            # root URI
            _initializeServer(
                it.server,
                params={
                    "initializationOptions": {
                        "project_file": p.join(TEST_PROJECT, "vimhdl.prj")
                    }
                },
            )

        @it.should("lint file when opening it")  # type: ignore
        def test():
            source = p.join(TEST_PROJECT, "basic_library", "clk_en_generator.vhd")

            it.assertCountEqual(checkLintFileOnOpen(source), [])


it.createTests(globals())



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=function-redefined
# pylint: disable=invalid-name
# pylint: disable=missing-docstring

import logging
import os.path as p
import pprint
import tempfile
import time
from contextlib import contextmanager
from multiprocessing import Event
from threading import Thread
from typing import Any, Iterator, List

import mock
import six

from nose2.tools import such  # type: ignore

from hdl_checker.tests import assertCountEqual, getTestTempPath, setupTestSuport

from hdl_checker.builder_utils import BuilderName
from hdl_checker.exceptions import UnknownParameterError
from hdl_checker.parsers.config_parser import ConfigParser
from hdl_checker.path import Path
from hdl_checker.types import BuildFlagScope
from hdl_checker.utils import ON_WINDOWS, toBytes

_logger = logging.getLogger(__name__)

TEST_TEMP_PATH = getTestTempPath(__name__)
TEST_PROJECT = p.join(TEST_TEMP_PATH, "test_project")

SOME_ABS_PATH = "C:\\some\\abs\\path.VHDL" if ON_WINDOWS else "/some/abs/path.VHDL"


@contextmanager
def fileWithContent(content):  # type: (bytes) -> Iterator[str]
    with tempfile.NamedTemporaryFile(delete=False) as fd:
        print("Writing to %s (%s)" % (fd, fd.name))
        fd.write(content)
        fd.flush()
        yield fd.name


such.unittest.TestCase.maxDiff = None

with such.A("config parser object") as it:

    if six.PY2:
        # Can't use assertCountEqual for lists of unhashable types.
        # Workaround for https://bugs.python.org/issue10242
        it.assertCountEqual = assertCountEqual(it)

    @it.has_setup
    def setup():
        # type: (...) -> Any
        setupTestSuport(TEST_TEMP_PATH)

    #  @it.has_teardown
    #  def teardown():
    #      for temp_path in ('.build', '.hdl_checker'):
    #          temp_path = p.abspath(p.join(CONFIG_PARSER_SUPPORT_PATH,
    #                                       temp_path))
    #          if p.exists(temp_path):
    #              shutil.rmtree(temp_path)

    @it.should(
        "raise UnknownParameterError exception when an unknown " "parameter is found"
    )
    def test_raises_exception():
        # type: (...) -> Any
        with it.assertRaises(UnknownParameterError):
            with fileWithContent(b"foo = bar") as name:
                parser = ConfigParser(Path(name))
                parser.parse()

    with it.having("a regular file"):

        @it.has_setup
        def setup():
            # type: (...) -> Any
            it.path = Path(tempfile.mktemp())

            contents = toBytes(
                """
single_build_flags[vhdl] = -single_build_flag_0
dependencies_build_flags[vhdl] = --vhdl-batch
global_build_flags[vhdl] = -globalvhdl -global-vhdl-flag

single_build_flags[verilog] = -single_build_flag_0
dependencies_build_flags[verilog] = --verilog-batch
global_build_flags[verilog] = -globalverilog -global-verilog-flag

single_build_flags[systemverilog] = -single_build_flag_0
dependencies_build_flags[systemverilog] = --systemverilog-batch
global_build_flags[systemverilog] = -globalsystemverilog -global-systemverilog-flag

builder = msim
target_dir = .build

vhdl work sample_file.vhd -sample_file_flag
vhdl work sample_package.vhdl -sample_package_flag
vhdl work TESTBENCH.VHD -build-in some way

vhdl lib {0}

verilog work foo.v -some-flag some value

systemverilog work bar.sv some sv flag
""".format(
                    SOME_ABS_PATH
                )
            )

            with open(it.path.name, "wb") as fd:
                fd.write(contents)
                fd.flush()

        @it.should("find the correct info")
        def test_parsing_regular_file():
            # type: (...) -> Any
            parser = ConfigParser(it.path)
            config = parser.parse()

            _logger.info("Parsed config:\n%s", pprint.pformat(config))
            sources = config.pop("sources")

            it.assertDictEqual(
                config,
                {
                    "builder_name": BuilderName.msim.name,
                    "vhdl": {
                        "flags": {
                            BuildFlagScope.all.value: (
                                "-globalvhdl",
                                "-global-vhdl-flag",
                            ),
                            BuildFlagScope.dependencies.value: ("--vhdl-batch",),
                            BuildFlagScope.single.value: ("-single_build_flag_0",),
                        }
                    },
                    "verilog": {
                        "flags": {
                            BuildFlagScope.all.value: (
                                "-globalverilog",
                                "-global-verilog-flag",
                            ),
                            BuildFlagScope.dependencies.value: ("--verilog-batch",),
                            BuildFlagScope.single.value: ("-single_build_flag_0",),
                        }
                    },
                    "systemverilog": {
                        "flags": {
                            BuildFlagScope.all.value: (
                                "-globalsystemverilog",
                                "-global-systemverilog-flag",
                            ),
                            BuildFlagScope.dependencies.value: (
                                "--systemverilog-batch",
                            ),
                            BuildFlagScope.single.value: ("-single_build_flag_0",),
                        }
                    },
                },
            )

            def _resolve(path):
                return p.join(it.path.dirname, path)

            it.assertCountEqual(
                sources,
                [
                    (
                        _resolve("sample_file.vhd"),
                        {"library": "work", "flags": ("-sample_file_flag",)},
                    ),
                    (
                        _resolve("sample_package.vhdl"),
                        {"library": "work", "flags": ("-sample_package_flag",)},
                    ),
                    (
                        _resolve("TESTBENCH.VHD"),
                        {"library": "work", "flags": ("-build-in", "some", "way")},
                    ),
                    (SOME_ABS_PATH, {"library": "lib", "flags": ()}),
                    (
                        _resolve("foo.v"),
                        {"library": "work", "flags": ("-some-flag", "some", "value")},
                    ),
                    (
                        _resolve("bar.sv"),
                        {"library": "work", "flags": ("some", "sv", "flag")},
                    ),
                ],
            )

        @it.should("only parse when the file actually changes")
        def test_only_parse_when_source_changes():
            # type: (...) -> Any
            parser = ConfigParser(it.path)

            _logger.info("Parsing %s for the 1st time", it.path)
            with mock.patch.object(parser, "_parseLine") as _parseLine:
                parser.parse()
                _parseLine.assert_called()

            time.sleep(0.1)

            _logger.info("Parsing %s for the 2nd time", it.path)
            with mock.patch.object(parser, "_parseLine") as _parseLine:
                parser.parse()
                _parseLine.assert_not_called()

        @it.should("report parsing in progress")
        def test_should_report_parsing_in_progress():
            # type: (...) -> Any
            parser = ConfigParser(it.path)
            it.assertFalse(parser.isParsing(), "Parser should not be busy right now")

            event = Event()

            # Make parsing wait until we set the event to mitigate race
            # conditions
            def _parse(*args, **kwargs):
                it.assertTrue(event.wait(timeout=3), "Timeout waiting for event")

            with mock.patch.object(parser, "_parse", _parse):
                thread = Thread(target=parser.parse)
                thread.start()
                time.sleep(0.1)
                it.assertTrue(
                    parser.isParsing(), "Parser should indicate it's still parsing"
                )
                event.set()


it.createTests(globals())



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"GHDL builder implementation"

import os
import os.path as p
import re
from glob import glob
from typing import Any, Iterable, List, Optional

from .base_builder import BaseBuilder

from hdl_checker.diagnostics import BuilderDiag, DiagType
from hdl_checker.parsers.elements.identifier import Identifier
from hdl_checker.path import Path
from hdl_checker.types import BuildFlags, FileType
from hdl_checker.utils import runShellCommand


class GHDL(BaseBuilder):
    """
    Builder implementation of the GHDL compiler
    """

    # Implementation of abstract class properties
    builder_name = "ghdl"
    file_types = {FileType.vhdl}

    # Default build flags
    default_flags = {"global": {FileType.vhdl: ("-fexplicit", "-frelaxed-rules")}}

    # GHDL specific class properties
    _stdout_message_parser = re.compile(
        r"^(?P<filename>.*):(?=\d)"
        r"(?P<line_number>\d+):"
        r"(?P<column_number>\d+):"
        r"((?P<is_warning>warning:)\s*|\s*)"
        r"(?P<error_message>.*)",
        re.I,
    ).finditer

    _scan_library_paths = re.compile(
        r"^\s*(actual prefix|library directory):" r"\s*(?P<library_path>.*)\s*"
    ).search

    _shouldIgnoreLine = re.compile(
        "|".join([r"^\s*$", r"ghdl: compilation error"])
    ).match

    _iter_rebuild_units = re.compile(
        r'((?P<unit_type>entity|package) "(?P<unit_name>\w+)" is obsoleted by (entity|package) "\w+"'
        r"|"
        r"file (?P<rebuild_path>.*)\s+has changed and must be reanalysed)",
        flags=re.I,
    ).finditer

    def __init__(self, *args, **kwargs):
        self._version = ""
        super(GHDL, self).__init__(*args, **kwargs)

    def _makeRecords(self, line):
        # type: (str) -> Iterable[BuilderDiag]
        for match in GHDL._stdout_message_parser(line):
            info = match.groupdict()
            diag = BuilderDiag(
                builder_name=self.builder_name, text=info.get("error_message", None)
            )

            if info["is_warning"]:
                diag.severity = DiagType.WARNING
            else:
                diag.severity = DiagType.ERROR

            filename = info.get("filename")
            line_number = info.get("line_number")
            column_number = info.get("column_number")

            if filename is not None:
                diag.filename = Path(filename)
            if line_number is not None:
                diag.line_number = line_number
            if column_number is not None:
                diag.column_number = column_number

            self._logger.info("Diag: %s", diag)
            yield diag

    def _checkEnvironment(self):
        stdout = runShellCommand(["ghdl", "--version"])
        self._version = re.findall(r"(?<=GHDL)\s+([^\s]+)\s+", stdout[0])[0]
        self._logger.info(
            "GHDL version string: '%s'. " "Version number is '%s'",
            stdout[:-1],
            self._version,
        )

    @staticmethod
    def isAvailable():
        try:
            runShellCommand(["ghdl", "--version"])
            return True
        except OSError:
            return False

    def _parseBuiltinLibraries(self):
        # type: (...) -> Any
        """
        Discovers libraries that exist regardless before we do anything
        """
        for line in runShellCommand(["ghdl", "--dispconfig"]):
            library_path_match = self._scan_library_paths(line)
            if library_path_match:
                library_path = library_path_match.groupdict()["library_path"]
                self._logger.debug("library path is %s", library_path)

                # Up to v0.36 ghdl kept libraries at
                #   <library_path>/<vhdl starndard>/<name>
                # but his has been changed to
                #   <library_path>/<name>/<vhdl starndard>
                libraries_paths = glob(
                    p.join(library_path, "v93", "*")
                    if self._version < "0.36"
                    else p.join(library_path, "*")
                )

                for path in filter(p.isdir, libraries_paths):
                    name = path.split(p.sep)[-1]
                    yield Identifier(name.strip(), case_sensitive=False)

    def _getGhdlArgs(self, path, library, flags=None):
        # type: (Path, Identifier, Optional[BuildFlags]) -> List[str]
        """
        Return the GHDL arguments that are common to most calls
        """
        cmd = [
            "-P%s" % self._work_folder,
            "--work=%s" % library,
            "--workdir=%s" % self._work_folder,
        ]
        if flags:
            cmd += flags
        cmd += [path.name]
        return cmd

    def _importSource(self, path, library, flags=None):
        """
        Runs GHDL with import source switch
        """
        vhdl_std = tuple(
            filter(lambda flag: flag.startswith("--std="), flags or tuple())
        )
        self._logger.debug("Importing source with std '%s'", vhdl_std)
        cmd = ["ghdl", "-i"] + self._getGhdlArgs(path, library, tuple(vhdl_std))
        return cmd

    def _analyzeSource(self, path, library, flags=None):
        # type: (Path, Identifier, Optional[BuildFlags]) -> List[str]
        """
        Runs GHDL with analyze source switch
        """
        return ["ghdl", "-a"] + self._getGhdlArgs(path, library, flags)

    def _checkSyntax(self, path, library, flags=None):
        # type: (Path, Identifier, Optional[BuildFlags]) -> List[str]
        """
        Runs GHDL with syntax check switch
        """
        return ["ghdl", "-s"] + self._getGhdlArgs(path, library, flags)

    def _buildSource(self, path, library, flags=None):
        # type: (Path, Identifier, Optional[BuildFlags]) -> Iterable[str]
        self._importSource(path, library, flags)

        stdout = []  # type: List[str]
        for cmd in (
            self._analyzeSource(path, library, flags),
            self._checkSyntax(path, library, flags),
        ):
            stdout += runShellCommand(cmd)

        return stdout

    def _createLibrary(self, _):
        workdir = p.join(self._work_folder)
        if not p.exists(workdir):
            os.makedirs(workdir)

    def _searchForRebuilds(self, line):
        rebuilds = []

        for match in self._iter_rebuild_units(line):
            mdict = match.groupdict()
            # When compilers reports units out of date, they do this
            # by either
            #  1. Giving the path to the file that needs to be rebuilt
            #     when sources are from different libraries
            #  2. Reporting which design unit has been affected by a
            #     given change.
            if "rebuild_path" in mdict and mdict["rebuild_path"] is not None:
                rebuilds.append(mdict)
            else:
                rebuilds.append(
                    {"unit_type": mdict["unit_type"], "unit_name": mdict["unit_name"]}
                )

        return rebuilds



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=function-redefined
# pylint: disable=invalid-name
# pylint: disable=missing-docstring
# pylint: disable=protected-access
# pylint: disable=useless-object-inheritance

import logging
import os.path as p

from mock import MagicMock, patch

from hdl_checker.tests import TestCase, disableVunit, getTestTempPath

from hdl_checker.builder_utils import (
    foundVunit,
    getBuilderByName,
    getVunitSources,
    getWorkingBuilders,
)
from hdl_checker.builders.fallback import Fallback
from hdl_checker.builders.ghdl import GHDL
from hdl_checker.builders.msim import MSim
from hdl_checker.builders.xvhdl import XVHDL
from hdl_checker.path import Path
from hdl_checker.types import FileType

_logger = logging.getLogger(__name__)

TEST_TEMP_PATH = getTestTempPath(__name__)


def _path(*args):
    # type: (str) -> str
    "Helper to reduce foorprint of p.join(TEST_TEMP_PATH, *args)"
    return p.join(TEST_TEMP_PATH, *args)


class TestBuilderUtils(TestCase):
    def test_getBuilderByName(self):
        self.assertEqual(getBuilderByName("msim"), MSim)
        self.assertEqual(getBuilderByName("xvhdl"), XVHDL)
        self.assertEqual(getBuilderByName("ghdl"), GHDL)
        self.assertEqual(getBuilderByName("other"), Fallback)

    def test_getWorkingBuilders(self):
        # Test no working builders
        _logger.info("Checking no builder works")
        self.assertFalse(list(getWorkingBuilders()))

        # Patch one builder
        with patch.object(MSim, "isAvailable", staticmethod(lambda: True)):
            _logger.info("Checking MSim works")
            self.assertCountEqual(list(getWorkingBuilders()), [MSim])
            # Patch another builder
            with patch.object(GHDL, "isAvailable", staticmethod(lambda: True)):
                _logger.info("Checking MSim and GHDL work")
                self.assertCountEqual(list(getWorkingBuilders()), [MSim, GHDL])


class Library(object):  # pylint: disable=too-few-public-methods
    def __init__(self, name):
        self._name = name

    @property
    def name(self):
        return self._name


class SourceFile(object):
    def __init__(self, name, library, vhdl_standard="2008"):
        self._name = name
        self._library = Library(library)
        self._vhdl_standard = vhdl_standard

    @property
    def name(self):
        return self._name

    @property
    def library(self):
        return self._library

    @property
    def vhdl_standard(self):
        return self._vhdl_standard


class TestGetVunitSources(TestCase):
    def test_vunit_not_found(self):
        builder = MagicMock()
        with disableVunit:
            self.assertFalse(list(getVunitSources(builder)))

    @patch("vunit.VUnit.get_source_files")
    def test_vhdl_builder(self, meth):
        meth.side_effect = [
            [
                SourceFile(name=_path("path_0.vhd"), library="libary_0"),
                SourceFile(name=_path("path_1.vhd"), library="libary_1"),
            ]
        ]

        builder = MagicMock()
        builder.builder_name = "msim"
        builder.file_types = {FileType.vhdl}

        self.maxDiff = None
        self.assertTrue(foundVunit(), "Need VUnit for this test")
        # Should only have VHDL files
        sources = list(getVunitSources(builder))

        meth.assert_called_once()

        self.assertCountEqual(
            sources,
            {
                (Path(_path("path_0.vhd")), "libary_0", ("-2008",)),
                (Path(_path("path_1.vhd")), "libary_1", ("-2008",)),
            },
        )

    @patch("vunit.verilog.VUnit.get_source_files")
    def test_systemverilog_only_builder(self, meth):
        meth.side_effect = [
            [
                SourceFile(name=_path("path_0.vhd"), library="libary_0"),
                SourceFile(name=_path("path_1.vhd"), library="libary_1"),
            ]
        ]

        builder = MagicMock()
        builder.builder_name = "msim"
        builder.file_types = {FileType.systemverilog}

        self.assertTrue(foundVunit(), "Need VUnit for this test")
        # Should only have VHDL files
        sources = list(getVunitSources(builder))

        meth.assert_called_once()

        self.assertCountEqual(
            sources,
            {
                (Path(_path("path_0.vhd")), "libary_0", ("-2008",)),
                (Path(_path("path_1.vhd")), "libary_1", ("-2008",)),
            },
        )

    @patch("hdl_checker.builder_utils._getSourcesFromVUnitModule")
    def test_verilog_only_builder(self, meth):
        builder = MagicMock()
        builder.builder_name = "msim"
        builder.file_types = {FileType.verilog}

        self.assertTrue(foundVunit(), "Need VUnit for this test")
        self.assertFalse(list(getVunitSources(builder)))
        meth.assert_not_called()

    @patch("vunit.VUnit.get_source_files")
    @patch("vunit.verilog.VUnit.get_source_files")
    def test_vhdl_and_systemverilog_only_builder(self, vhdl_method, sv_method):
        vhdl_method.side_effect = [
            [
                SourceFile(name=_path("path_0.vhd"), library="libary_0"),
                SourceFile(name=_path("path_1.vhd"), library="libary_1"),
            ]
        ]

        sv_method.side_effect = [
            [
                SourceFile(name=_path("path_2.sv"), library="libary_2"),
                SourceFile(name=_path("path_3.sv"), library="libary_3"),
            ]
        ]

        builder = MagicMock()
        builder.builder_name = "xvhdl"
        builder.file_types = {FileType.vhdl, FileType.systemverilog}

        self.assertTrue(foundVunit(), "Need VUnit for this test")
        # Should only have VHDL files
        sources = list(getVunitSources(builder))

        vhdl_method.assert_called_once()
        sv_method.assert_called_once()

        self.assertCountEqual(
            sources,
            {
                (Path(_path("path_0.vhd")), "libary_0", ()),
                (Path(_path("path_1.vhd")), "libary_1", ()),
                (Path(_path("path_2.sv")), "libary_2", ()),
                (Path(_path("path_3.sv")), "libary_3", ()),
            },
        )



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.

import logging
from typing import Any, FrozenSet, Iterable, Optional, Set, Tuple

from hdl_checker import types as t  # pylint: disable=unused-import
from hdl_checker.utils import HashableByKey

_logger = logging.getLogger(__name__)

Location = Tuple[Any, Any]
LocationList = Iterable[Location]


class ParsedElement(HashableByKey):
    def __init__(self, locations=None):
        # type: (Optional[LocationList]) -> None
        set_of_locations = set()  # type: Set[Location]
        for line_number, column_number in locations or []:
            set_of_locations.add(
                (
                    None if line_number is None else int(line_number),
                    None if column_number is None else int(column_number),
                )
            )

        self._locations = frozenset(set_of_locations)

    @property
    def locations(self):
        return self._locations

    @property
    def __hash_key__(self):
        return (self.locations,)



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"ModelSim builder implementation"

import os
import os.path as p
import re
from shutil import copyfile
from typing import Any, Iterable, List, Optional

from .base_builder import BaseBuilder

from hdl_checker.database import Database
from hdl_checker.diagnostics import BuilderDiag, DiagType
from hdl_checker.parsers.elements.identifier import Identifier
from hdl_checker.path import Path
from hdl_checker.types import BuildFlags, BuildFlagScope, FileType
from hdl_checker.utils import runShellCommand


class MSim(BaseBuilder):
    """Builder implementation of the ModelSim compiler"""

    # Implementation of abstract class properties
    builder_name = "msim"
    file_types = {FileType.vhdl, FileType.verilog, FileType.systemverilog}

    # MSim specific class properties
    _stdout_message_scanner = re.compile(
        r"""^\*\*\s*
                (?P<severity>[WE])\w+\s*
                (:?\(suppressible\))?:\s*
                (:?
                    (:?\s*\[\d+\])?\s*
                    (?P<filename>.*(?=\(\d+\)))
                    \((?P<line_number>\d+)\):
                |
                    \(vcom-\d+\)
                )?
            \s*(?P<error_message>.*)\s*""",
        flags=re.VERBOSE,
    ).finditer

    _should_ignore = re.compile(
        "|".join(
            [
                r"^\s*$",
                r"^(?!\*\*\s(Error|Warning)\b).*",
                r".*VHDL Compiler exiting\s*$",
            ]
        )
    ).match

    _iter_rebuild_units = re.compile(
        r"("
        r"Recompile\s*(?P<lib_name_0>\w+)\.(?P<unit_name_0>\w+)\s+because"
        r"\s+.*?\s+ha(?:ve|s) changed"
        r"|"
        r"^\*\* Warning:.*\(vcom-1127\)\s*Entity\s(?P<lib_name_1>\w+)\."
        r"(?P<unit_name_1>\w+).*"
        r")"
    ).finditer

    _BuilderLibraryScanner = re.compile(
        r"^\"(?P<library_name>\w+)\""
        r"\s+maps to directory\s+"
        r"(?P<library_path>.*)\.$",
        re.I,
    )

    # Default build flags
    default_flags = {
        BuildFlagScope.dependencies: {
            FileType.vhdl: ("-defercheck", "-nocheck", "-permissive"),
            FileType.verilog: ("-permissive",),
            FileType.systemverilog: ("-permissive",),
        },
        BuildFlagScope.single: {
            FileType.vhdl: (
                "-check_synthesis",
                "-lint",
                "-rangecheck",
                "-pedanticerrors",
            ),
            FileType.verilog: ("-lint", "-hazards", "-pedanticerrors"),
            FileType.systemverilog: ("-lint", "-hazards", "-pedanticerrors"),
        },
        BuildFlagScope.all: {
            FileType.vhdl: ("-explicit",),
            FileType.verilog: (),
            FileType.systemverilog: (),
        },
    }

    def _shouldIgnoreLine(self, line):
        return self._should_ignore(line)

    def __init__(self, work_folder, database):
        # type: (Path, Database) -> None
        self._version = ""
        self._modelsim_ini = Path(p.join(work_folder.name, "modelsim.ini"))
        super(MSim, self).__init__(work_folder, database)

    def setup(self):
        # type: (...) -> Any
        super(MSim, self).setup()
        if not self._iniFileExists():
            self._createIniFile()

    def _makeRecords(self, line):
        # type: (str) -> Iterable[BuilderDiag]
        for match in self._stdout_message_scanner(line):  # type: ignore
            info = match.groupdict()

            self._logger.debug("Parsed dict: %s", repr(info))

            text = re.sub(
                r"\s*\((vcom|vlog)-\d+\)\s*", " ", info["error_message"]
            ).strip()

            diag = BuilderDiag(builder_name=self.builder_name, text=text)

            error_code = None

            if ("vcom-" in line) or ("vlog" in line):
                error_code = re.findall(r"((?:vcom-|vlog-)\d+)", line)[0]

            diag.error_code = error_code

            filename = info.get("filename")
            line_number = info.get("line_number")
            column_number = info.get("column_number")

            if filename is not None:
                diag.filename = Path(filename)
            if line_number is not None:
                diag.line_number = line_number
            if column_number is not None:
                diag.column_number = column_number

            if info.get("severity", None) in ("W", "e"):
                diag.severity = DiagType.WARNING
            elif info.get("severity", None) in ("E", "e"):
                diag.severity = DiagType.ERROR

            yield diag

    def _checkEnvironment(self):
        stdout = runShellCommand(["vcom", "-version"])
        self._version = re.findall(r"(?<=vcom)\s+([\w\.]+)\s+(?=Compiler)", stdout[0])[
            0
        ]
        self._logger.debug(
            "vcom version string: '%s'. " "Version number is '%s'",
            stdout,
            self._version,
        )

    @staticmethod
    def isAvailable():
        try:
            runShellCommand(["vcom", "-version"])
            runShellCommand(["vlog", "-version"])
            return True
        except OSError:
            return False

    def _parseBuiltinLibraries(self):
        # type: (...) -> Any
        "Discovers libraries that exist regardless before we do anything"
        if not self._iniFileExists():
            self._createIniFile()
        for line in runShellCommand(["vmap"]):
            for match in self._BuilderLibraryScanner.finditer(line):
                yield Identifier(match.groupdict()["library_name"], False)

    def _searchForRebuilds(self, line):
        rebuilds = []
        for match in self._iter_rebuild_units(line):
            mdict = match.groupdict()
            library_name = mdict["lib_name_0"] or mdict["lib_name_1"]
            unit_name = mdict["unit_name_0"] or mdict["unit_name_1"]
            if None not in (library_name, unit_name):
                rebuilds.append({"library_name": library_name, "unit_name": unit_name})
            else:  # pragma: no cover
                _msg = "Something wrong while parsing '%s'. " "Match is '%s'" % (
                    line,
                    mdict,
                )
                self._logger.error(_msg)
                assert 0, _msg

        return rebuilds

    def _buildSource(self, path, library, flags=None):
        # type: (Path, Identifier, Optional[BuildFlags]) -> Iterable[str]
        filetype = FileType.fromPath(path)
        if filetype == FileType.vhdl:
            return self._buildVhdl(path, library, flags)
        if filetype in (FileType.verilog, FileType.systemverilog):
            return self._buildVerilog(path, library, flags)

        self._logger.error(  # pragma: no cover
            "Unknown file type %s for path '%s'", filetype, path
        )
        return ""

    def _getExtraFlags(self, lang):
        # type: (FileType) -> Iterable[str]
        """
        Gets extra flags configured for the specific language
        """
        libs = []  # type: List[str]
        for library in self._added_libraries | self._external_libraries[lang]:
            libs = ["-L", library.name]
        for path in self._include_paths[lang]:
            libs += ["+incdir+" + str(path)]
        return libs

    def _buildVhdl(self, path, library, flags=None):
        # type: (Path, Identifier, Optional[BuildFlags]) -> Iterable[str]
        "Builds a VHDL file"
        assert isinstance(library, Identifier)
        cmd = [
            "vcom",
            "-modelsimini",
            self._modelsim_ini.name,
            "-quiet",
            "-work",
            p.join(self._work_folder, library.name),
        ]
        if flags:  # pragma: no cover
            cmd += flags
        cmd += [path.name]

        return runShellCommand(cmd)

    def _buildVerilog(self, path, library, flags=None):
        # type: (Path, Identifier, Optional[BuildFlags]) -> Iterable[str]
        "Builds a Verilog/SystemVerilog file"
        cmd = [
            "vlog",
            "-modelsimini",
            self._modelsim_ini.name,
            "-quiet",
            "-work",
            p.join(self._work_folder, library.name),
        ]

        if FileType.fromPath(path) == FileType.systemverilog:
            cmd += ["-sv"]
        if flags:  # pragma: no cover
            cmd += flags

        cmd += self._getExtraFlags(FileType.verilog)
        cmd += [path.name]

        return runShellCommand(cmd)

    def _createLibrary(self, library):
        if p.exists(p.join(self._work_folder, library.name)):
            self._logger.debug("Path for library '%s' already exists", library)
            return
        self._mapLibrary(library)
        self._logger.debug("Added and mapped library '%s'", library)

    def _iniFileExists(self):
        # type: (...) -> bool
        """
        Checks if the modelsim.ini file exists at the expected location
        """
        return p.exists(self._modelsim_ini.abspath)

    def _createIniFile(self):
        # type: (...) -> Any
        """
        Adds a library to a non-existent ModelSim init file
        """
        if not p.exists(self._work_folder):  # pragma: no cover
            os.makedirs(self._work_folder)

        self._logger.debug("Creating modelsim.ini at '%s'", self._modelsim_ini)

        modelsim_env = os.environ.get("MODELSIM")
        if modelsim_env is not None:  # pragma: no cover
            self._logger.info(
                "MODELSIM environment variable set to %s, using "
                "this path as default modelsim.ini",
                modelsim_env,
            )
            # Copy the modelsim.ini as indicated by the MODELSIM environment
            # variable
            copyfile(modelsim_env, self._modelsim_ini.abspath)
        else:
            runShellCommand(["vmap", "-c"], cwd=self._work_folder)

    def deleteLibrary(self, library):
        "Deletes a library from ModelSim init file"
        if not p.exists(p.join(self._work_folder, library)):
            self._logger.warning("Library %s doesn't exists", library)
            return None
        return runShellCommand(
            ["vdel", "-modelsimini", self._modelsim_ini, "-lib", library, "-all"]
        )

    def _mapLibrary(self, library):
        # type: (Identifier) -> None
        """
        Adds a library to an existing ModelSim init file
        """
        runShellCommand(["vlib", p.join(self._work_folder, library.name)])

        runShellCommand(
            [
                "vmap",
                "-modelsimini",
                self._modelsim_ini.name,
                library.name,
                p.join(self._work_folder, library.name),
            ]
        )



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=function-redefined
# pylint: disable=missing-docstring
# pylint: disable=protected-access
# pylint: disable=invalid-name

import logging
import os
import os.path as p
import shutil

from webtest import TestApp  # type: ignore

from hdl_checker.tests import TestCase, getTestTempPath, setupTestSuport

import hdl_checker.handlers as handlers
from hdl_checker.builders.fallback import Fallback
from hdl_checker.builders.ghdl import GHDL
from hdl_checker.builders.msim import MSim
from hdl_checker.builders.xvhdl import XVHDL
from hdl_checker.config_generators.simple_finder import SimpleFinder

try:  # Python 3.x
    import unittest.mock as mock
except ImportError:  # Python 2.x
    import mock  # type: ignore


TEST_TEMP_PATH = getTestTempPath(__name__)
TEST_PROJECT = p.abspath(p.join(TEST_TEMP_PATH, "test_project"))

SERVER_LOG_LEVEL = os.environ.get("SERVER_LOG_LEVEL", "INFO")

_logger = logging.getLogger(__name__)
HDL_CHECKER_BASE_PATH = p.abspath(p.join(p.dirname(__file__), "..", ".."))

BUILDER_CLASS_MAP = {"msim": MSim, "xvhdl": XVHDL, "ghdl": GHDL, "fallback": Fallback}


class TestConfigGenerator(TestCase):
    maxDiff = None

    @classmethod
    def setUpClass(cls):
        setupTestSuport(TEST_TEMP_PATH)

    def setUp(self):
        self.app = TestApp(handlers.app)

        # Needs to agree with vroom test file
        self.dummy_test_path = p.join(TEST_TEMP_PATH, "dummy_test_path")

        self.assertFalse(
            p.exists(self.dummy_test_path),
            "Path '%s' shouldn't exist right now" % p.abspath(self.dummy_test_path),
        )

        os.makedirs(self.dummy_test_path)

        os.mkdir(p.join(self.dummy_test_path, "path_a"))
        os.mkdir(p.join(self.dummy_test_path, "path_b"))
        os.mkdir(p.join(self.dummy_test_path, "v_includes"))
        os.mkdir(p.join(self.dummy_test_path, "sv_includes"))
        # Create empty sources and some extra files as well
        for path in (
            "README.txt",  # This shouldn't be included
            "nonreadable.txt",  # This shouldn't be included
            p.join("path_a", "some_source.vhd"),
            p.join("path_a", "header_out_of_place.vh"),
            p.join("path_a", "source_tb.vhd"),
            p.join("path_b", "some_source.vhd"),
            p.join("path_b", "a_verilog_source.v"),
            p.join("path_b", "a_systemverilog_source.sv"),
            # Create headers for both extensions
            p.join("v_includes", "verilog_header.vh"),
            p.join("sv_includes", "systemverilog_header.svh"),
            # Make the tree 'dirty' with other source types
            p.join("path_a", "not_hdl_source.log"),
            p.join("path_a", "not_hdl_source.py"),
        ):
            _logger.info("Writing to %s", path)
            open(p.join(self.dummy_test_path, path), "w").write("")

    def teardown(self):
        # Create a dummy arrangement of sources
        if p.exists(self.dummy_test_path):
            _logger.info("Removing %s", repr(self.dummy_test_path))
            shutil.rmtree(self.dummy_test_path)

    @mock.patch(
        "hdl_checker.config_generators.simple_finder.isFileReadable",
        lambda path: "nonreadable" not in path.name,
    )
    def test_run_simple_config_gen(self):
        # type: (...) -> None
        finder = SimpleFinder([self.dummy_test_path])

        config = finder.generate()

        self.assertCountEqual(
            config.pop("sources"),
            {
                p.join(self.dummy_test_path, "path_a", "some_source.vhd"),
                p.join(self.dummy_test_path, "path_b", "a_systemverilog_source.sv"),
                p.join(self.dummy_test_path, "path_a", "source_tb.vhd"),
                p.join(self.dummy_test_path, "path_b", "some_source.vhd"),
                p.join(self.dummy_test_path, "path_b", "a_verilog_source.v"),
            },
        )

        self.assertCountEqual(
            config.pop("systemverilog"),
            {"include_paths": {p.join(self.dummy_test_path, "sv_includes")}},
        )

        self.assertCountEqual(
            config.pop("verilog"),
            {
                "include_paths": {
                    p.join(self.dummy_test_path, "path_a"),
                    p.join(self.dummy_test_path, "v_includes"),
                }
            },
        )

        self.assertFalse(config)



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=function-redefined, missing-docstring, protected-access

import json
import logging
import os

import six

from nose2.tools import such  # type: ignore

from hdl_checker.tests import assertCountEqual, writeListToFile

from hdl_checker.parsers.elements.design_unit import DesignUnitType
from hdl_checker.parsers.verilog_parser import VerilogDesignUnit, VerilogParser
from hdl_checker.path import Path
from hdl_checker.serialization import StateEncoder, jsonObjectHook

_logger = logging.getLogger(__name__)

_FILENAME = "source.v"

with such.A("Verilog source file object") as it:
    if six.PY2:
        it.assertCountEqual = assertCountEqual(it)

    with it.having("a module code"):

        @it.has_setup
        def setup():
            if os.path.exists(_FILENAME):
                os.remove(_FILENAME)
            it._code = """
module clock_divider
    #(parameter DIVISION = 5)
    (// Usual ports
    input clk,
    input rst,
    // Output clock divided
    output       clk_div);
""".splitlines()

            writeListToFile(_FILENAME, it._code)

        @it.has_teardown
        def teardown():
            if os.path.exists(_FILENAME):
                os.remove(_FILENAME)

        @it.should("parse a file without errors")
        def test():
            it.source = VerilogParser(Path(_FILENAME))

        @it.should("return its design units")  # type: ignore
        def test():
            design_units = list(it.source.getDesignUnits())
            _logger.debug("Design units: %s", design_units)
            it.assertCountEqual(
                design_units,
                [
                    VerilogDesignUnit(
                        owner=it.source.filename,
                        name="clock_divider",
                        type_=DesignUnitType.entity,
                        locations={(0, None)},
                    )
                ],
            )

        @it.should("return no dependencies")  # type: ignore
        def test():
            it.assertFalse(it.source.getDependencies())

        @it.should("return source modification time")  # type: ignore
        def test():
            it.assertEqual(os.path.getmtime(_FILENAME), it.source.getmtime())

        @it.should("return no libraries")  # type: ignore
        def test():
            it.assertFalse(it.source.getLibraries())

        @it.should("report as equal after recovering from cache")  # type:ignore
        def test():
            state = json.dumps(it.source, cls=StateEncoder)
            _logger.info("State before: %s", state)
            recovered = json.loads(state, object_hook=jsonObjectHook)
            it.assertEqual(it.source.filename, recovered.filename)

    with it.having("a package code"):

        @it.has_setup
        def setup():
            if os.path.exists(_FILENAME):
                os.remove(_FILENAME)
            it._code = """
package msgPkg;
  integer  errCnt  = 0;
  integer  warnCnt = 0;
endpackage
""".splitlines()

            writeListToFile(_FILENAME, it._code)

        @it.has_teardown
        def teardown():
            if os.path.exists(_FILENAME):
                os.remove(_FILENAME)

        @it.should("return its design units")  # type: ignore
        def test():
            design_units = list(it.source.getDesignUnits())
            _logger.debug("Design units: %s", design_units)
            it.assertCountEqual(
                design_units,
                [
                    VerilogDesignUnit(
                        owner=it.source.filename,
                        name="msgPkg",
                        type_=DesignUnitType.package,
                        locations={(0, None)},
                    )
                ],
            )

        @it.should("return no libraries")  # type: ignore
        def test():
            it.assertEqual(it.source.getLibraries(), [])


it.createTests(globals())



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"Base class for creating a project file"

import os
import os.path as p
from typing import Iterable, List

from .base_generator import BaseGenerator

from hdl_checker.path import Path
from hdl_checker.types import FileType, UnknownTypeExtension
from hdl_checker.utils import isFileReadable

_SOURCE_EXTENSIONS = "vhdl", "sv", "v"
_HEADER_EXTENSIONS = "vh", "svh"


class SimpleFinder(BaseGenerator):
    """
    Implementation of BaseGenerator that searches for paths on a given
    set of paths recursively
    """

    def __init__(self, paths):  # type: (List[str]) -> None
        super(SimpleFinder, self).__init__()
        self._logger.debug("Search paths: %s", paths)
        self._paths = {Path(x) for x in paths}
        self._valid_extensions = tuple(
            list(_SOURCE_EXTENSIONS) + list(_HEADER_EXTENSIONS)
        )

    def _getLibrary(self, path):  # type: (Path) -> str
        """
        Returns the library name given the path. On this implementation this
        returns a default name; child classes can override this to provide
        specific names (say the library name is embedded on the path itself or
        on the file's contents)
        """
        return NotImplemented

    def _findSources(self):
        # type: (...) -> Iterable[Path]
        """
        Iterates over the paths and searches for relevant files by extension.
        """
        for path in self._paths:
            for dirpath, _, filenames in os.walk(path.name):
                for filename in filenames:
                    full_path = Path(p.join(dirpath, filename))

                    if not p.isfile(full_path.name):
                        continue

                    try:
                        # FileType.fromPath will fail if the file's extension is not
                        # valid (one of '.vhd', '.vhdl', '.v', '.vh', '.sv',
                        # '.svh')
                        FileType.fromPath(full_path)
                    except UnknownTypeExtension:
                        continue

                    if isFileReadable(full_path):
                        yield full_path

    def _populate(self):  # type: (...) -> None
        for path in self._findSources():
            library = self._getLibrary(path)
            self._addSource(
                path, library=None if library is NotImplemented else library
            )



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"Language server protocol implementation"

import logging
import os.path as p
import tempfile
from typing import Any, Dict, Iterable, Optional, Set

import six
from pyls import lsp as defines  # type: ignore
from pyls._utils import debounce  # type: ignore
from pyls.python_ls import PythonLanguageServer  # type: ignore
from pyls.uris import to_fs_path  # type: ignore
from pyls.workspace import Workspace  # type: ignore

from hdl_checker import DEFAULT_PROJECT_FILE
from hdl_checker.config_generators.simple_finder import SimpleFinder
from hdl_checker.diagnostics import CheckerDiagnostic, DiagType
from hdl_checker.exceptions import UnknownParameterError
from hdl_checker.hdl_checker_base import HdlCodeCheckerBase
from hdl_checker.path import Path
from hdl_checker.utils import logCalls

_logger = logging.getLogger(__name__)

LINT_DEBOUNCE_S = 0.5  # 500 ms

URI = str

if six.PY2:
    FileNotFoundError = (  # pylint: disable=redefined-builtin,invalid-name
        IOError,
        OSError,
    )


def checkerDiagToLspDict(diag):
    # type: (...) -> Any
    """
    Converts a CheckerDiagnostic object into the dictionary with into the LSP
    expects
    """
    _logger.debug(diag)

    # Translate the error into LSP severity
    severity = diag.severity

    if severity in (DiagType.INFO, DiagType.STYLE_INFO):
        severity = defines.DiagnosticSeverity.Hint
    elif severity in (DiagType.STYLE_WARNING, DiagType.STYLE_ERROR):
        severity = defines.DiagnosticSeverity.Information
    elif severity in (DiagType.WARNING,):
        severity = defines.DiagnosticSeverity.Warning
    elif severity in (DiagType.ERROR,):
        severity = defines.DiagnosticSeverity.Error
    else:
        severity = defines.DiagnosticSeverity.Error

    result = {
        "source": diag.checker,
        "range": {
            "start": {
                "line": (diag.line_number or 1) - 1,
                "character": (diag.column_number or 1) - 1,
            },
            "end": {
                "line": (diag.line_number or 1) - 1,
                "character": (diag.column_number or 1) - 1,
            },
        },
        "message": diag.text,
        "severity": severity,
    }

    if diag.error_code:
        result["code"] = diag.error_code

    return result


class HdlCodeCheckerServer(HdlCodeCheckerBase):
    """
    HDL Checker project builder class
    """

    def __init__(self, workspace, root_path):
        # type: (Workspace, Optional[str]) -> None
        self._workspace = workspace
        if root_path is None:
            root_path = tempfile.mkdtemp(prefix="hdl_checker_")

        super(HdlCodeCheckerServer, self).__init__(Path(root_path))

    def _handleUiInfo(self, message):
        # type: (...) -> Any
        _logger.debug("UI info: %s (workspace=%s)", message, self._workspace)
        if self._workspace: # pragma: no cover
            self._workspace.show_message(message, defines.MessageType.Info)

    def _handleUiWarning(self, message):
        # type: (...) -> Any
        _logger.debug("UI warning: %s (workspace=%s)", message, self._workspace)
        if self._workspace: # pragma: no cover
            self._workspace.show_message(message, defines.MessageType.Warning)

    def _handleUiError(self, message):
        # type: (...) -> Any
        _logger.debug("UI error: %s (workspace=%s)", message, self._workspace)
        if self._workspace: # pragma: no cover
            self._workspace.show_message(message, defines.MessageType.Error)


class HdlCheckerLanguageServer(PythonLanguageServer):
    """ Implementation of the Microsoft VSCode Language Server Protocol
    https://github.com/Microsoft/language-server-protocol/blob/master/versions/protocol-1-x.md
    """

    # pylint: disable=too-many-public-methods,redefined-builtin

    def __init__(self, *args, **kwargs):
        # type: (...) -> None
        self._checker = None  # type: Optional[HdlCodeCheckerServer]
        super(HdlCheckerLanguageServer, self).__init__(*args, **kwargs)
        # Default checker
        self._onConfigUpdate({"project_file": None})
        self._global_diags = set()  # type: Set[CheckerDiagnostic]
        self._initialization_options = {}  # type: Dict[str, Any]

    @logCalls
    def capabilities(self):
        # type: (...) -> Any
        "Returns language server capabilities"
        return {"textDocumentSync": defines.TextDocumentSyncKind.FULL}

    @logCalls
    def m_initialized(self, **_kwargs):
        """
        Enables processing of actions that were generated upon m_initialize and
        were delayed because the client might need further info (for example to
        handle window/showMessage requests)
        """
        self._onConfigUpdate(self._initialization_options)
        return super(HdlCheckerLanguageServer, self).m_initialized(**_kwargs)

    @logCalls
    def m_initialize(
        self,
        processId=None,
        rootUri=None,
        rootPath=None,
        initializationOptions=None,
        **_kwargs
    ):
        # type: (...) -> Any
        """
        Initializes the language server
        """
        result = super(HdlCheckerLanguageServer, self).m_initialize(
            processId=processId,
            rootUri=rootUri,
            rootPath=rootPath,
            initializationOptions={},
        )

        self._initialization_options = initializationOptions

        return result

    def _onConfigUpdate(self, options):
        # type: (...) -> Any
        """
        Updates the checker server from options if the 'project_file' key is
        present. Please not that this is run from both initialize and
        workspace/did_change_configuration and when ran initialize the LSP
        client might not ready to take messages. To circumvent this, make sure
        m_initialize returns before calling this to actually configure the
        server.
        """
        root_path = None

        if self.workspace and self.workspace.root_uri is not None:
            root_path = to_fs_path(self.workspace.root_uri)

        self._checker = HdlCodeCheckerServer(self.workspace, root_path=root_path)

        _logger.debug("Updating from %s", options)

        # Clear previous diagnostics
        self._global_diags = set()

        path = self._getProjectFilePath(options)

        if path is not None:
            try:
                self._checker.setConfig(path)
                return
            except UnknownParameterError as exc:
                _logger.info("Failed to read config from %s: %s", path, exc)
                return
            except FileNotFoundError:
                # If the file couldn't be found, proceed to searching the root
                # URI (if it has been set)
                pass

        if not self.workspace or not self.workspace.root_path:
            _logger.debug("No workspace or root path not set, can't search files")
            return

        # Having no project file but with root URI triggers searching for
        # sources automatically
        config = SimpleFinder([self.workspace.root_path]).generate()
        self.workspace.show_message(
            "Added {} files from {}".format(
                len(config["sources"]), self.workspace.root_path
            ),
            defines.MessageType.Info,
        )
        self._checker.configure(config)

    def _getProjectFilePath(self, options=None):
        # type: (...) -> Optional[str]
        """
        Tries to get 'project_file' from the options dict and combine it with
        the root URI as provided by the workspace
        """
        path = (options or {}).get("project_file", DEFAULT_PROJECT_FILE)

        # Path has been explicitly set to none
        if path is None:
            return None

        # Project file will be related to the root path
        if self.workspace:
            path = p.join(self.workspace.root_path or "", path)

        return path

    @debounce(LINT_DEBOUNCE_S, keyed_by="doc_uri")
    def lint(self, doc_uri, is_saved):
        # type: (...) -> Any
        diagnostics = set(self._getDiags(doc_uri, is_saved)) | self._global_diags

        # Since we're debounced, the document may no longer be open
        if doc_uri in self.workspace.documents:
            # Both checker methods return generators, convert to a list before
            # returning
            self.workspace.publish_diagnostics(
                doc_uri, list([checkerDiagToLspDict(x) for x in diagnostics])
            )

    def _getDiags(self, doc_uri, is_saved):
        # type: (URI, bool) -> Iterable[CheckerDiagnostic]
        """
        Gets diags of the URI, wether from the saved file or from its
        contents
        """
        if self._checker is None: # pragma: no cover
            _logger.debug("No checker, won't try to get diagnostics")
            return ()

        # If the file has not been saved, use the appropriate method, which
        # will involve dumping the modified contents into a temporary file
        path = Path(to_fs_path(doc_uri))

        _logger.info("Linting %s (saved=%s)", repr(path), is_saved)

        if is_saved:
            diags = self._checker.getMessagesByPath(path)
        else:
            text = self.workspace.get_document(doc_uri).source
            diags = self._checker.getMessagesWithText(path, text)

        # LSP diagnostics are only valid for the scope of the resource and
        # hdl_checker may return a tree of issues, so need to filter those out
        return (diag for diag in diags if diag.filename in (path, None))

    @logCalls
    def m_workspace__did_change_configuration(self, settings=None):
        # type: (...) -> Any
        self._onConfigUpdate(settings or {})



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"Diagnostics holders for checkers"

import os.path as p
from typing import Optional

from hdl_checker.path import Path
from hdl_checker.utils import HashableByKey, samefile

# pylint: disable=useless-object-inheritance

CHECKER_NAME = "HDL Checker"
STATIC_CHECKER_NAME = "HDL Checker/static"


class DiagType(object):  # pylint: disable=too-few-public-methods
    """
    Enum-like class for error types
    """

    NONE = "None"
    INFO = "Info"
    WARNING = "Warning"
    ERROR = "Error"
    STYLE_INFO = "Info (style)"
    STYLE_WARNING = "Warning (style)"
    STYLE_ERROR = "Error (style)"


class CheckerDiagnostic(HashableByKey):  # pylint: disable=too-many-instance-attributes
    """
    Base container for diagnostics
    """

    def __init__(
        self,  # pylint: disable=too-many-arguments
        checker,  # type: str
        text,  # type: str
        filename=None,  # type: Optional[Path]
        line_number=None,
        column_number=None,
        error_code=None,
        severity=None,
    ):

        # Checker can't be changed
        self.__checker = CHECKER_NAME if checker is None else checker

        # Modifiable attributes
        self.filename = filename  # type: Optional[Path]
        self.error_code = error_code
        self.text = str(text)

        # Modifiable with rules
        self.__line_number = None if line_number is None else int(line_number)
        self.__column = None if column_number is None else int(column_number)
        self.__severity = severity

        if line_number is not None:
            self.line_number = line_number
        if column_number is not None:
            self.column_number = column_number
        if severity is not None:
            self.severity = severity

    def __repr__(self):
        return (
            '{}(checker="{}", filename={}, line_number={}, column_number={}, '
            "error_code={}, severity={}, text={})".format(
                self.__class__.__name__,
                repr(self.checker),
                repr(self.filename),
                repr(self.line_number),
                repr(self.column_number),
                repr(self.error_code),
                repr(self.severity),
                repr(self.text),
            )
        )

    @property
    def __hash_key__(self):
        return (
            self.filename,
            self.checker,
            self.column_number,
            self.error_code,
            self.line_number,
            self.severity,
            self.text,
        )

    #  def __hash__(self):
    #      # Don't include the filename in the hash key to avoid it being used on
    #      # __eq__
    #      return hash((self.filename.abspath, super(CheckerDiagnostic, self).__hash__()))

    #  def __eq__(self, other):
    #      # type: (...) -> bool
    #      if (
    #          self.filename is not None
    #          and p.exists(self.filename.name)
    #          and other.filename is not None
    #          and p.exists(other.filename.name)
    #      ):
    #          if not samefile(self.filename, other.filename):
    #              return False

    #      return super(CheckerDiagnostic, self).__eq__(other)

    def toDict(self):
        """Returns a dict representation of the object. All keys are always
        present but not values are necessearily set, in which case their values
        will be 'None'. Dict has the folowwing format:

            - checker: (string) Checker name
            - filename: (string)
            - error_code: (string)
            - text: (string)
            - line_number: (int or None)
            - column_number: (int or None)
            - severity: (string) Values taken from DiagType
        """
        return {
            "checker": self.checker,
            "filename": str(self.filename),
            "error_code": self.error_code,
            "text": self.text,
            "line_number": self.line_number,
            "column_number": self.column_number,
            "severity": self.severity,
        }

    @classmethod
    def fromDict(cls, state):
        "Creates a diagnostics objects from state dict"
        return cls(
            checker=state["checker"],
            filename=state["filename"],
            error_code=state["error_code"],
            text=state["text"],
            line_number=state["line_number"],
            column_number=state["column_number"],
            severity=state["severity"],
        )

    @property
    def checker(self):
        "Full checker name"
        return self.__checker

    @property
    def line_number(self):
        "Diagnostics line number"
        if self.__line_number is not None:
            return int(self.__line_number)
        return self.__line_number

    @line_number.setter
    def line_number(self, value):
        self.__line_number = int(value)

    @property
    def column_number(self):
        "Diagnostics column_number"
        if self.__column is not None:
            return int(self.__column)
        return self.__column

    @column_number.setter
    def column_number(self, value):
        self.__column = int(value)

    @property
    def severity(self):
        "Diagnostics severity (use diagnostics.DiagType for consistency)"
        return self.__severity

    @severity.setter
    def severity(self, value):
        assert value in DiagType.__dict__.values(), "Invalid severity {}".format(
            repr(value)
        )
        self.__severity = value


class PathNotInProjectFile(CheckerDiagnostic):
    """
    Reports a check request on a file whose path in not present on the project
    file
    """

    def __init__(self, path):
        super(PathNotInProjectFile, self).__init__(
            checker=CHECKER_NAME,
            filename=path,
            severity=DiagType.WARNING,
            text='Path "{}" not found in project file'.format(path),
        )


class StaticCheckerDiag(CheckerDiagnostic):
    "Base diagnostics issues from static checks"

    def __init__(
        self,  # pylint: disable=too-many-arguments
        text,
        severity,
        filename=None,
        line_number=None,
        column_number=None,
        error_code=None,
    ):

        assert severity in (
            DiagType.STYLE_INFO,
            DiagType.STYLE_WARNING,
            DiagType.STYLE_ERROR,
        ), "Static checker diags should only carry style error types"

        super(StaticCheckerDiag, self).__init__(
            checker=STATIC_CHECKER_NAME,
            text=text,
            severity=severity,
            filename=filename,
            line_number=line_number,
            column_number=column_number,
            error_code=error_code,
        )


class LibraryShouldBeOmited(StaticCheckerDiag):
    "Library declaration should be ommited"

    def __init__(self, library, filename=None, line_number=None, column_number=None):
        super(LibraryShouldBeOmited, self).__init__(
            line_number=line_number,
            column_number=column_number,
            filename=filename,
            severity=DiagType.STYLE_INFO,
            text="Declaration of library '{library}' can be omitted".format(
                library=library
            ),
        )


class ObjectIsNeverUsed(StaticCheckerDiag):
    "Reports an object that was created but never used"

    def __init__(
        self,  # pylint: disable=too-many-arguments
        filename=None,
        line_number=None,
        column_number=None,
        object_name=None,
        object_type=None,
    ):
        super(ObjectIsNeverUsed, self).__init__(
            filename=filename,
            line_number=line_number,
            column_number=column_number,
            severity=DiagType.STYLE_WARNING,
            text="{} '{}' is never used".format(
                str(object_type).capitalize(), object_name
            ),
        )


class BuilderDiag(CheckerDiagnostic):
    """
    Reports issues when checking a file whose path in not present on the
    project file
    """

    def __init__(
        self,  # pylint: disable=too-many-arguments
        builder_name,
        text,
        filename=None,
        line_number=None,
        column_number=None,
        error_code=None,
        severity=None,
    ):
        super(BuilderDiag, self).__init__(
            checker="{}/{}".format(CHECKER_NAME, builder_name),
            text=text,
            filename=filename,
            severity=severity,
            line_number=line_number,
            column_number=column_number,
            error_code=error_code,
        )

    #  def __hash__(self):
    #      #  for item in ('checker', 'column_number', 'error_code',
    #      #          'filename', 'line_number', 'severity', 'text'):
    #      #      try:
    #      #          hash(getattr(self, item))
    #      #      except TypeError:
    #      #          print("item %s is not hashable!" % item)
    #      return super(CheckerDiagnostic, self).__hash__()
    #      #  assert False, 'hey'


class FailedToCreateProject(CheckerDiagnostic):
    """
    Reports problems when reading the project file
    """

    def __init__(self, exception):
        text = "Exception while creating server: {}"

        super(FailedToCreateProject, self).__init__(
            checker=None, severity=DiagType.ERROR, text=text.format(str(exception))
        )


class DependencyNotUnique(CheckerDiagnostic):
    """
    Searching for a dependency should yield a single source file
    """

    def __init__(
        self, filename, design_unit, choices, line_number=None, column_number=None
    ):
        _choices = list(choices)
        text = (
            "Dependency '{}' (library={}) has {} definitions (files are {}). "
            "The selected option may not be the correct one".format(
                design_unit.name,
                design_unit.library,
                len(_choices),
                ", ".join(('"%s"' % x for x in _choices)),
            )
        )

        super(DependencyNotUnique, self).__init__(
            checker=None,
            filename=filename,
            severity=DiagType.STYLE_WARNING,
            line_number=line_number,
            column_number=column_number,
            text=text,
        )



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"Project wide database"

# pylint: disable=useless-object-inheritance

import logging
import os.path as p
from itertools import chain
from threading import RLock
from typing import (
    Any,
    Dict,
    FrozenSet,
    Iterable,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    Union,
)

from hdl_checker.diagnostics import (
    CheckerDiagnostic,
    DependencyNotUnique,
    PathNotInProjectFile,
)
from hdl_checker.parser_utils import flattenConfig, getSourceParserFromPath
from hdl_checker.parsers.elements.dependency_spec import DependencySpec
from hdl_checker.parsers.elements.design_unit import tAnyDesignUnit
from hdl_checker.parsers.elements.identifier import Identifier
from hdl_checker.path import Path, TemporaryPath
from hdl_checker.types import BuildFlags, BuildFlagScope, FileType
from hdl_checker.utils import HashableByKey, getMostCommonItem, isFileReadable

try:
    from functools import lru_cache
except ImportError:
    from backports.functools_lru_cache import lru_cache  # type: ignore

_logger = logging.getLogger(__name__)
_DEFAULT_LIBRARY_NAME = Identifier("library")

UnresolvedLibrary = Union[Identifier, None]
LibraryUnitTuple = Tuple[UnresolvedLibrary, Identifier]


class Database(HashableByKey):
    "Stores info on and provides operations for a project file set"

    def __init__(self):  # type: () -> None
        self._lock = RLock()

        self._paths = set()  # type: Set[Path]
        self._parse_timestamp = {}  # type: Dict[Path, float]
        self._library_map = {}  # type: Dict[Path, Identifier]
        self._flags_map = {}  # type: Dict[Path, Dict[BuildFlagScope, BuildFlags]]
        self._dependencies_map = {}  # type: Dict[Path, Set[DependencySpec]]
        self._inferred_libraries = set()  # type: Set[Path]
        self._design_units = set()  # type: Set[tAnyDesignUnit]
        self._diags = {}  # type: Dict[Path, Set[CheckerDiagnostic]]

        # Use this to know which methods should be cache
        self._cached_methods = {
            getattr(self, x)
            for x in dir(self)
            if hasattr(getattr(self, x), "cache_clear")
        }

    @property
    def __hash_key__(self):
        return 0

    @property
    def design_units(self):  # type: (...) -> FrozenSet[tAnyDesignUnit]
        "Set of design units found"
        return frozenset(self._design_units)

    def refresh(self):
        # type: (...) -> Any
        """
        Clears caches, inferred libraries and parses and checks if any source
        should be parsed
        """
        self._clearLruCaches()

        while self._inferred_libraries:
            try:
                del self._library_map[self._inferred_libraries.pop()]
            except KeyError:  # pragma: no cover
                pass

        for path in self.paths:
            self._parseSourceIfNeeded(path)

    def configure(self, root_config, root_path):
        # type: (Dict[str, Any], str) -> None
        """
        Handles adding sources, libraries and flags from a dict
        """
        for entry in flattenConfig(root_config, root_path):
            self.addSource(
                path=entry.path,
                library=entry.library,
                single_flags=entry.single_flags,
                dependencies_flags=entry.dependencies_flags,
            )

    def addSource(self, path, library, single_flags=None, dependencies_flags=None):
        # type: (Path, Optional[str], Optional[BuildFlags], Optional[BuildFlags]) -> None
        """
        Adds a source to the database, triggering its parsing even if the
        source has already been added previously
        """
        _logger.info(
            "Adding %s, library=%s, flags=(single=%s, dependencies=%s)",
            path,
            library,
            single_flags,
            dependencies_flags,
        )
        self._paths.add(path)
        self._flags_map[path] = {
            BuildFlagScope.single: tuple(single_flags or ()),
            BuildFlagScope.dependencies: tuple(dependencies_flags or ()),
        }
        if library is not None:
            self._library_map[path] = Identifier(
                library, case_sensitive=FileType.fromPath(path) != FileType.vhdl
            )

        # TODO: Parse on a process pool
        self._parseSource(path)

    def removeSource(self, path):
        # type: (Path) -> None
        """
        Removes a path from the database. No error is raised if the path wasn't
        added previously. In this case, avoid clearning LRU caches
        """
        _logger.info("Removing %s from database", path)
        clear_lru_caches = False

        with self._lock:
            units = frozenset(self._getDesignUnitsByPath(path))
            self._design_units -= units

            if units:
                clear_lru_caches = True

            try:
                self._paths.remove(path)
                clear_lru_caches = True
            except KeyError:
                pass

            try:
                del self._parse_timestamp[path]
                clear_lru_caches = True
            except KeyError:
                pass

            try:
                del self._library_map[path]
                clear_lru_caches = True
            except KeyError:
                pass

            try:
                del self._flags_map[path]
                clear_lru_caches = True
            except KeyError:
                pass

            try:
                del self._dependencies_map[path]
                clear_lru_caches = True
            except KeyError:
                pass

            try:
                del self._diags[path]
                clear_lru_caches = True
            except KeyError:
                pass

            if clear_lru_caches:
                self._clearLruCaches()

    def _addDiagnostic(self, diagnostic):
        # type: (CheckerDiagnostic) -> None
        """
        Adds a diagnostic to the diagnostic map. Diagnostics can then be read
        to report processing internals and might make it to the user interface
        """
        _logger.debug("Adding diagnostic %s", diagnostic)
        assert diagnostic.filename is not None
        if diagnostic.filename not in self._diags:
            self._diags[diagnostic.filename] = set()

        self._diags[diagnostic.filename].add(diagnostic)

    def getDiagnosticsForPath(self, path):
        # type: (Path) -> Iterable[CheckerDiagnostic]
        """
        Returns diagnostics generated a path. It does not trigger any
        processing or analysis though
        """
        return self._diags.get(path, ())

    def __jsonEncode__(self):
        """
        Gets a dict that describes the current state of this object
        """
        state = {"sources": []}

        for path in self._paths:
            source_info = {
                "path": path,
                "mtime": self._parse_timestamp[path],
                "flags": {
                    BuildFlagScope.single.value: self._flags_map[path].get(
                        BuildFlagScope.single, ()
                    ),
                    BuildFlagScope.dependencies.value: self._flags_map[path].get(
                        BuildFlagScope.dependencies, ()
                    ),
                },
                "dependencies": tuple(self._dependencies_map.get(path, ())),
                "diags": tuple(),
            }

            library = self._library_map.get(path, None)
            if library is not None:
                source_info["library"] = library

            state["sources"].append(source_info)

        state["inferred_libraries"] = tuple(self._inferred_libraries)
        state["design_units"] = tuple(self._design_units)

        return state

    @classmethod
    def __jsonDecode__(cls, state):
        # pylint: disable=protected-access
        obj = cls()
        obj._design_units = {x for x in state.pop("design_units")}
        obj._inferred_libraries = {x for x in state.pop("inferred_libraries")}
        for info in state.pop("sources"):
            path = info.pop("path")
            obj._paths.add(path)
            obj._parse_timestamp[path] = float(info.pop("mtime"))

            if "library" in info:
                obj._library_map[path] = info.pop("library")

            obj._flags_map[path] = {}
            obj._flags_map[path][BuildFlagScope.single] = tuple(
                info.get("flags", {}).pop(BuildFlagScope.single.value, ())
            )
            obj._flags_map[path][BuildFlagScope.dependencies] = tuple(
                info.pop("flags", {}).pop(BuildFlagScope.dependencies.value, ())
            )
            obj._dependencies_map[path] = {x for x in info.pop("dependencies")}
            obj._diags[path] = {x for x in info.pop("diags")}
        # pylint: enable=protected-access

        return obj

    def getFlags(self, path, scope=None):
        # type: (Path, Optional[BuildFlagScope]) -> BuildFlags
        """
        Return a list of flags for the given path or an empty tuple if the path
        is not found in the database.
        """
        return self._flags_map.get(path, {}).get(scope or BuildFlagScope.single, ())

    @property
    def paths(self):
        # type: () -> Iterable[Path]
        "Returns a list of paths currently in the database"
        return frozenset(self._paths)

    def _updatePathLibrary(self, path, library):
        # type: (Path, Identifier) -> None
        """
        Updates dependencies of the given path so they reflect the change in
        their owner's path
        """

        if path not in self._library_map:
            _logger.info("Setting library for '%s' to '%s'", path, library)
        else:
            current_library = self._library_map.get(path)
            # No change, avoid manipulating the database
            if current_library == library:
                return

            _logger.info(
                "Replacing old library '%s' for '%s' with '%s'",
                current_library,
                path,
                library,
            )

        self._library_map[path] = library

        # Nothing to resolve if the dependency entry is empty
        if not self._dependencies_map.get(path, None):
            return

        # Extract the unresolved dependencies that will be replaced
        unresolved_dependencies = {
            x for x in self._dependencies_map[path] if x.library is None
        }

        # DependencySpec is not mutable, so we actually need to replace the objects
        for dependency in unresolved_dependencies:
            self._dependencies_map[path].add(
                DependencySpec(
                    owner=dependency.owner,
                    name=dependency.name,
                    library=library,
                    locations=dependency.locations,
                )
            )

        # Safe to remove the unresolved ones
        self._dependencies_map[path] -= unresolved_dependencies

    @lru_cache()
    def getLibrary(self, path):
        # type: (Path) -> UnresolvedLibrary
        "Gets a library of a given source (this is likely to be removed)"
        self._parseSourceIfNeeded(path)
        if path not in self.paths:
            # Add the path to the project but put it on a different library
            self._parseSourceIfNeeded(path)
            self._updatePathLibrary(path, Identifier("not_in_project", True))
            # If path is not on the list of paths added, report this. If the
            # config is valid
            if not isinstance(path, TemporaryPath):
                self._addDiagnostic(PathNotInProjectFile(path))

        elif path not in self._library_map:
            # Library is not defined, try to infer
            _logger.info("Library for '%s' not set, inferring it", path)
            library = self._inferLibraryIfNeeded(path)
            if library is not None:
                self._updatePathLibrary(path, library)

        return self._library_map.get(path, None)

    def _parseSourceIfNeeded(self, path):
        # type: (Path) -> None
        """
        Parses a given path if needed, removing info from the database prior to that
        """
        if not isFileReadable(path):
            _logger.warning("Won't parse file that's not readable %s", repr(path))
            return

        # Sources will get parsed on demand
        mtime = p.getmtime(path.name)

        if mtime == self._parse_timestamp.get(path, 0):
            return

        self._parseSource(path)

    def _parseSource(self, path):
        # type: (Path) -> None
        """
        Extracts info from a source, taking care of removing previously defined
        items before
        """
        _logger.debug("Parsing %s", path)

        src_parser = getSourceParserFromPath(path)
        design_units = src_parser.getDesignUnits()
        dependencies = src_parser.getDependencies()

        with self._lock:
            # Update the timestamp
            self._parse_timestamp[path] = p.getmtime(str(path))

            # Remove all design units that referred to this path before adding new
            # ones, but use the non API method for that to avoid recursing
            self._design_units -= frozenset(self._getDesignUnitsByPath(path))

            self._design_units |= design_units
            self._dependencies_map[path] = dependencies
            self._clearLruCaches()

    def _clearLruCaches(self):
        "Clear caches from lru_caches"
        for meth in self._cached_methods:
            meth.cache_clear()

    def getDesignUnitsByPath(self, path):  # type: (Path) -> Set[tAnyDesignUnit]
        "Gets the design units for the given path (if any)"
        self._parseSourceIfNeeded(path)
        return self._getDesignUnitsByPath(path)

    @lru_cache(maxsize=128, typed=False)
    def _getDesignUnitsByPath(self, path):  # type: (Path) -> Set[tAnyDesignUnit]
        """
        Gets the design units for the given path (if any). Differs from the
        public method in that changes to the file are not checked before
        running.
        """
        return {x for x in self.design_units if x.owner == path}

    def getDependenciesByPath(self, path):
        # type: (Path) -> FrozenSet[DependencySpec]
        """
        Returns parsed dependencies for the given path
        """
        self._parseSourceIfNeeded(path)
        if path not in self._dependencies_map:
            return frozenset()
        return frozenset(self._dependencies_map[path])

    def getPathsByDesignUnit(self, unit):
        # type: (tAnyDesignUnit) -> Iterator[Path]
        """
        Return the source (or sources) that define the given design
        unit
        """
        return (
            design_unit.owner
            for design_unit in self.design_units
            if (unit.name, unit.type_) == (design_unit.name, design_unit.type_)
        )

    def _inferLibraryIfNeeded(self, path):
        # type: (Path) -> UnresolvedLibrary
        """
        Tries to infer which library the given path should be compiled on by
        looking at where and how the design units it defines are used
        """
        # Find all units this path defines
        units = set(self.getDesignUnitsByPath(path))
        _logger.debug("Units defined here in %s: %s", path, list(map(str, units)))
        # Store all cases to use in case there are multiple libraries that
        # could be used. If that happens, we'll use the most common one
        all_libraries = list(
            chain.from_iterable(
                self.getLibrariesReferredByUnit(name=unit.name) for unit in units
            )
        )

        libraries = set(all_libraries)

        if not libraries:
            _logger.info("Couldn't work out a library for path %s", path)
            library = None
        elif len(libraries) == 1:
            library = libraries.pop()
        else:
            library = getMostCommonItem(all_libraries)
            _msg = []
            for lib in libraries:
                _msg.append("%s (x%d)" % (lib, all_libraries.count(lib)))
            _logger.info(
                "Path %s is in %d libraries: %s, using %s",
                path,
                len(libraries),
                ", ".join(_msg),
                library,
            )

        self._inferred_libraries.add(path)
        return library

    @lru_cache()
    def getLibrariesReferredByUnit(self, name, library=None):
        # type: (Identifier, UnresolvedLibrary) -> List[Identifier]
        """
        Gets libraries that the (library, name) pair is used throughout the
        project
        """
        result = []  # List[Identifier]
        for path, dependencies in self._dependencies_map.items():
            for dependency in dependencies:
                library = dependency.library
                if library is None or name != dependency.name:
                    continue

                # If the dependency's library refers to 'work', it's actually
                # referring to the library its owner is in
                if library is None:
                    library = self._library_map.get(path, None)
                if library is not None:
                    result.append(library)

        return result

    @lru_cache()
    def getPathsDefining(self, name, library=None):
        # type: (Identifier, UnresolvedLibrary) -> Iterable[Path]
        """
        Search for paths that define a given name optionally inside a library.
        """
        units = {unit for unit in self.design_units if unit.name == name}

        if not units:
            _logger.debug(
                "Could not find any source defining name=%s, library=%s", name, library
            )
            return ()

        if library is not None:
            units_matching_library = {
                unit for unit in units if (self.getLibrary(unit.owner) == library)
            }

            if not units_matching_library:
                # If no units match when using the library, it means the database
                # is incomplete and we should try to infer the library from the
                # usage of this unit
                for owner in {x.owner for x in units}:
                    # Force getting library for this path to trigger library
                    # inference if needed
                    self.getLibrary(owner)
            else:
                units = units_matching_library

        # These include paths and temporary paths, which is what this method
        # should return but useless when reporting dependencies not unique.
        paths = {unit.owner for unit in units}

        _logger.debug(
            "There's %d path(s) defining %s.%s: %s", len(paths), library, name, paths
        )

        if len(paths) > 1:
            self._reportDependencyNotUnique(library=library, name=name, choices=paths)

        return paths

    def _reportDependencyNotUnique(self, library, name, choices):
        # type: (Optional[Identifier], Identifier, Iterable[Path]) -> None
        """
        Reports a dependency failed to be resolved due to multiple files
        defining the required design unit
        """
        # Filter out of choices paths that are temporary. If that reduces the
        # choices to 1 element, there's no need to report anything
        choices = {x for x in choices if not isinstance(x, TemporaryPath)}
        if len(choices) < 2:
            return

        # Reverse dependency search. Need to evaluate how this performs, but
        # we'll try to avoid creating a set to store DependencySpec objects for
        # now

        for dependency in (
            dependency
            for dependency in chain.from_iterable(self._dependencies_map.values())
            if (library, name) == (dependency.library, dependency.name)
        ):

            # Fill in a report for every occurence found
            for location in dependency.locations:
                self._addDiagnostic(
                    DependencyNotUnique(
                        filename=dependency.owner,
                        line_number=location[0],
                        column_number=location[1],
                        design_unit=dependency,
                        choices=choices,
                    )
                )

    def getDependenciesUnits(self, path):
        # type: (Path) -> Set[LibraryUnitTuple]
        """
        Returns design units that should be compiled before compiling the given
        path but only within the project file set. If a design unit can't be
        found in any source, it will be silently ignored.
        """
        # These paths define the dependencies that the original path has. In
        # the ideal case, each dependency is defined once and the config file
        # specifies the correct library, in which case we don't add any extra
        # warning.
        #
        # If a dependency is defined multiple times, issue a warning indicating
        # which one is going to actually be used and which are the other
        # options, just like what has been already implemented.
        #
        # If the library is not set for the a given path, try to guess it by
        # (1) given every design unit defined in this file, (2) search for
        # every file that also depends on it and (3) identify which library is
        # used. If all of them converge on the same library name, just use
        # that. If there's no agreement, use the library that satisfies the
        # path in question but warn the user that something is not right
        self._parseSourceIfNeeded(path)

        units = set()  # type: Set[LibraryUnitTuple]

        search_paths = set((path,))
        own_units = {
            (self.getLibrary(path), x.name) for x in self.getDesignUnitsByPath(path)
        }

        while search_paths:
            # Get the dependencies of the search paths and which design units
            # they define and remove the ones we've already seen
            new_deps = {
                (
                    dependency.library or self.getLibrary(dependency.owner),
                    dependency.name,
                )
                for search_path in search_paths
                for dependency in self._dependencies_map[search_path]
            } - units

            _logger.debug(
                "Searching %s resulted in dependencies: %s", search_paths, new_deps
            )

            # Add the new ones to the set tracking the dependencies seen since
            # the search started
            units |= new_deps

            # Paths to be searched on the next iteration are the paths of the
            # dependencies we have not seen before
            search_paths = set()

            for library, name in new_deps:
                new_paths = set(self.getPathsDefining(name=name, library=library))
                search_paths |= new_paths

            _logger.debug("Search paths: %s", search_paths)

        # Remove units defined by the path passed as argument
        units -= own_units
        return units

    def getBuildSequence(self, path, builtin_libraries=None):
        # type: (Path, Optional[Iterable[Identifier]]) -> Iterable[Tuple[Identifier, Path]]
        """
        Gets the build sequence that satisfies the preconditions to compile the
        given path
        """
        self._diags[path] = set()

        units_compiled = set()  # type: Set[LibraryUnitTuple]
        builtin_libraries = frozenset(builtin_libraries or [])

        units_to_build = self.getDependenciesUnits(path)
        paths_to_build = set(
            chain.from_iterable(
                self.getPathsDefining(name=name, library=library)
                for library, name in units_to_build
                if library not in builtin_libraries
            )
        )

        # Limit the number of iterations to the worst case of every pass
        # compiling only a single source and all of them having a chain of
        # dependencies on the previous one
        iteration_limit = len(paths_to_build) + 1

        for i in range(iteration_limit):
            paths_built = set()  # type: Set[Path]

            for current_path in paths_to_build:
                current_path_library = self.getLibrary(current_path)
                own = {
                    (current_path_library, x.name)
                    for x in self.getDesignUnitsByPath(current_path)
                }

                # Filter out dependencies that are on the 'use foo.all'
                # because this only indicates that a source needs a given
                # library to exist (which is handled by the builder).
                # Also filter out dependencies that are provided natively by
                # the builder.
                deps = {
                    (x.library or self.getLibrary(x.owner), x.name)
                    for x in self._dependencies_map[current_path]
                    if x.name != Identifier("all", False)
                    and x.library not in builtin_libraries
                }

                # Units still needed are the ones we haven't seen before
                still_needed = deps - units_compiled - own
                new_units = own - units_compiled

                if still_needed:
                    if _logger.isEnabledFor(logging.DEBUG):  # pragma: no cover
                        _msg = [(library, name.name) for library, name in still_needed]
                        _logger.debug("%s still needs %s", current_path, _msg)
                elif not new_units:
                    # If the current path only defines units that have been
                    # already compiled, skip it
                    _logger.debug("Path %s has nothing to add, skipping", current_path)
                    paths_built.add(current_path)
                else:
                    _logger.debug(
                        "Compiling %s adds %d new units: %s",
                        current_path,
                        len(new_units),
                        new_units,
                    )
                    yield self.getLibrary(
                        current_path
                    ) or _DEFAULT_LIBRARY_NAME, current_path
                    paths_built.add(current_path)
                    units_compiled |= own

            paths_to_build -= paths_built
            units_to_build -= units_compiled

            if not paths_built:
                if paths_to_build:
                    _logger.warning(
                        "%d paths were not built: %s",
                        len(paths_to_build),
                        list(map(str, paths_to_build)),
                    )
                else:
                    _logger.info("Nothing more to do after %d steps", i)
                return

            _logger.debug(
                "Got %d units compiled: %s", len(units_compiled), units_compiled
            )

        _logger.error("Iteration limit of %d reached", iteration_limit)



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"VHDL source file parser"

import logging
import re
from typing import Any, Dict, Generator, Optional, Set, Tuple, Union

from .elements.dependency_spec import DependencySpec
from .elements.design_unit import DesignUnitType, VhdlDesignUnit
from .elements.identifier import VhdlIdentifier

from hdl_checker.parsers.base_parser import BaseSourceFile
from hdl_checker.utils import readFile

_logger = logging.getLogger(__name__)

# Design unit scanner
_DESIGN_UNIT_SCANNER = re.compile(
    "|".join(
        [
            r"\bpackage\s+(?P<package_name>\w+)\s+is\b",
            r"\bentity\s+(?P<entity_name>\w+)\s+is\b",
            r"\blibrary\s+(?P<library_name>[\w,\s]+)\b",
            r"\bcontext\s+(?P<context_name>\w+)\s+is\b",
        ]
    ),
    flags=re.MULTILINE | re.IGNORECASE,
)

_LIBRARY_SCANNER = re.compile(
    r"library\s+([a-z]\w*(?:\s*,\s*[a-z]\w*){0,})\s*;",
    flags=re.MULTILINE | re.IGNORECASE,
)

_ADDITIONAL_DEPS_SCANNER = re.compile(
    r"\bpackage\s+body\s+(?P<package_body_name>\w+)\s+is\b",
    flags=re.MULTILINE | re.IGNORECASE,
)

IncompleteDependency = Dict[str, Union[str, Set[Any]]]


class VhdlParser(BaseSourceFile):
    """
    Parses and stores information about a source file such as design
    units it depends on and design units it provides
    """

    _comment = re.compile(r"--[^\n\r]*", flags=re.S)

    def _getSourceContent(self):
        # type: (...) -> Any
        """
        Replace everything from comment ('--') until a line break
        """
        content = readFile(str(self.filename))

        return self._comment.sub("", content)

    def _iterDesignUnitMatches(self):
        # type: (...) -> Any
        """
        Iterates over the matches of _DESIGN_UNIT_SCANNER against
        source's lines
        """
        content = self.getSourceContent()

        for match in _DESIGN_UNIT_SCANNER.finditer(content):
            start = match.start()
            yield match.groupdict(), content[:start].count("\n")

    def _getDependencies(self):  # type: () -> Generator[DependencySpec, None, None]
        lib_deps_regex = re.compile(
            r"|".join([r"%s\.\w+" % x for x in set(self.getLibraries() + ["work"])]),
            flags=re.I,
        )

        dependencies = {}  # type: ignore

        text = self.getSourceContent()

        for match in lib_deps_regex.finditer(text):

            library, unit = match.group().split(".")[
                :2
            ]  # type: Tuple[Optional[str], str]

            line_number = text[: match.end()].count("\n")
            column_number = len(text[: match.start()].split("\n")[-1])

            key = hash((library, unit))

            if key not in dependencies:
                dependencies[key] = {
                    "library": library,
                    "name": unit,
                    "locations": set(),
                }

            dependency = dependencies[key]
            dependency["locations"].add((line_number + 1, column_number + 1))

        # Done parsing, won't add any more locations, so generate the specs
        for dep in dependencies.values():
            # Remove references to 'work' (will treat library=None as work,
            # which also means not set in case of packages)
            if dep["library"].lower() == "work":
                dep_library = None
            else:
                dep_library = VhdlIdentifier(dep["library"])
            yield DependencySpec(
                owner=self.filename,
                name=VhdlIdentifier(dep["name"]),
                library=dep_library,
                locations=dep["locations"],
            )

        for match in _ADDITIONAL_DEPS_SCANNER.finditer(self.getSourceContent()):
            package_body_name = match.groupdict()["package_body_name"]
            line_number = int(text[: match.end()].count("\n"))
            column_number = len(text[: match.start()].split("\n")[-1])

            yield DependencySpec(
                owner=self.filename,
                name=VhdlIdentifier(package_body_name),
                library=None,
                locations={(line_number + 1, column_number + 1)},
            )

    def _getLibraries(self):
        # type: (...) -> Any
        """
        Parses the source file to find design units and dependencies
        """
        libs = set()  # type: Set[str]

        for match in _LIBRARY_SCANNER.finditer(self.getSourceContent()):
            for group in match.groups():
                libs = libs.union(set(map(str.strip, str(group).split(","))))

        # Replace references of 'work' for the actual library name
        if "work" in libs:
            libs = libs - {"work"}

        return libs

    def _getDesignUnits(self):  # type: () -> Generator[VhdlDesignUnit, None, None]
        """
        Parses the source file to find design units and dependencies
        """

        for match, line_number in self._iterDesignUnitMatches():
            locations = {(line_number, None)}

            if match["package_name"] is not None:
                yield VhdlDesignUnit(
                    owner=self.filename,
                    name=match["package_name"],
                    type_=DesignUnitType.package,
                    locations=locations,
                )

            elif match["entity_name"] is not None:
                yield VhdlDesignUnit(
                    owner=self.filename,
                    name=match["entity_name"],
                    type_=DesignUnitType.entity,
                    locations=locations,
                )
            elif match["context_name"] is not None:
                yield VhdlDesignUnit(
                    owner=self.filename,
                    name=match["context_name"],
                    type_=DesignUnitType.context,
                    locations=locations,
                )



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"VHDL, Verilog or SystemVerilog identifier and comparisons between them"

# pylint: disable=useless-object-inheritance


class Identifier(object):
    """
    VHDL, Verilog or SystemVerilog identifier to make it easier to handle case
    and comparisons between them
    """

    def __init__(self, name, case_sensitive=False):
        # type: (str, bool) -> None
        self.case_sensitive = case_sensitive
        self._display_name = str(name)
        self._name = self._display_name.lower()

    @property
    def name(self):
        "Normalized identifier name"
        return self._name

    @property
    def display_name(self):
        "Identifier name as given when creating the object"
        return self._display_name

    def __jsonEncode__(self):
        return {"name": self.display_name, "case_sensitive": self.case_sensitive}

    @classmethod
    def __jsonDecode__(cls, state):
        return cls(name=state.pop("name"), case_sensitive=state.pop("case_sensitive"))

    def __hash__(self):
        return hash(self.name)

    def __str__(self):
        return self.display_name

    def __repr__(self):
        if self.name == self.display_name:
            return "{}({})".format(self.__class__.__name__, repr(self.name))
        return "{}({}, display_name={})".format(
            self.__class__.__name__, repr(self.name), repr(self.display_name)
        )

    def __eq__(self, other):
        """Overrides the default implementation"""

        try:
            if self.case_sensitive and other.case_sensitive:
                return self.display_name == other.display_name
            return self.name == other.name
        except AttributeError:
            pass

        return NotImplemented  # pragma: no cover

    def __ne__(self, other):  # pragma: no cover
        """Overrides the default implementation (unnecessary in Python 3)"""
        result = self.__eq__(other)

        if result is not NotImplemented:
            return not result

        return NotImplemented


class VhdlIdentifier(Identifier):
    "Equivalent of Identifier(name, case_sensitive=False)"

    def __init__(self, name):
        # type: (str, ) -> None
        super(VhdlIdentifier, self).__init__(name=name, case_sensitive=False)

    @classmethod
    def __jsonDecode__(cls, state):
        return cls(name=state.pop("name"))


class VerilogIdentifier(Identifier):
    "Equivalent of Identifier(name, case_sensitive=True)"

    def __init__(self, name):
        # type: (str, ) -> None
        super(VerilogIdentifier, self).__init__(name=name, case_sensitive=True)

    @classmethod
    def __jsonDecode__(cls, state):
        return cls(name=state.pop("name"))



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"Misc tests of files, such as licensing and copyright"

# pylint: disable=function-redefined, missing-docstring, protected-access

import logging
import os.path as p
import re
import subprocess as subp

import parameterized  # type: ignore
import unittest2  # type: ignore

from hdl_checker.builder_utils import BuilderName, getBuilderByName
from hdl_checker.builders.fallback import Fallback
from hdl_checker.builders.ghdl import GHDL
from hdl_checker.builders.msim import MSim
from hdl_checker.builders.xvhdl import XVHDL
from hdl_checker.utils import readFile

_logger = logging.getLogger(__name__)

_HEADER = re.compile(
    r"(?:--|#) This file is part of HDL Checker\.\n"
    r"(?:--|#)\n"
    r"(?:--|#) Copyright \(c\) 2015 - 2019 suoto \(Andre Souto\)\n"
    r"(?:--|#)\n"
    r"(?:--|#) HDL Checker is free software: you can redistribute it and/or modify\n"
    r"(?:--|#) it under the terms of the GNU General Public License as published by\n"
    r"(?:--|#) the Free Software Foundation, either version 3 of the License, or\n"
    r"(?:--|#) \(at your option\) any later version\.\n"
    r"(?:--|#)\n"
    r"(?:--|#) HDL Checker is distributed in the hope that it will be useful,\n"
    r"(?:--|#) but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
    r"(?:--|#) MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE\.  See the\n"
    r"(?:--|#) GNU General Public License for more details\.\n"
    r"(?:--|#)\n"
    r"(?:--|#) You should have received a copy of the GNU General Public License\n"
    r"(?:--|#) along with HDL Checker\.  If not, see <http://www\.gnu\.org/licenses/>\.\n"
)


def _getFiles():
    for filename in subp.check_output(
        ["git", "ls-tree", "--name-only", "-r", "HEAD"]
    ).splitlines():
        yield p.abspath(filename).decode()


def _getRelevantFiles():
    def _fileFilter(path):
        # Exclude versioneer files
        if p.basename(path) in ("_version.py", "versioneer.py"):
            return False
        if p.join(".ci", "test_support") in path:
            return False
        return path.split(".")[-1] in ("py", "sh", "ps1")

    return filter(_fileFilter, _getFiles())


def checkFile(filename):
    lines = readFile(filename)

    match = _HEADER.search(lines)
    return match is not None


class TestFileHeaders(unittest2.TestCase):
    @parameterized.parameterized.expand([(x,) for x in _getRelevantFiles()])
    def test_has_license(self, path):
        self.assertTrue(checkFile(path))


class TestBuilderUtils(unittest2.TestCase):
    def test_getBuilderByName(self):
        self.assertEqual(getBuilderByName(BuilderName.msim.value), MSim)
        self.assertEqual(getBuilderByName(BuilderName.ghdl.value), GHDL)
        self.assertEqual(getBuilderByName(BuilderName.xvhdl.value), XVHDL)
        self.assertEqual(getBuilderByName("foo"), Fallback)



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"Top of the hdl_checker.parsers submodule"

# pylint: disable=useless-object-inheritance

import json
import logging
import os.path as p
from collections import namedtuple
from multiprocessing.pool import ThreadPool as Pool
from typing import Any, Dict, Iterable, Set, Tuple, Type, Union

import six

from .parsers.verilog_parser import VerilogParser
from .parsers.vhdl_parser import VhdlParser

from hdl_checker.path import Path
from hdl_checker.types import BuildFlagScope, FileType

_logger = logging.getLogger(__name__)

tSourceFile = Union[VhdlParser, VerilogParser]

if six.PY3:
    JSONDecodeError = json.decoder.JSONDecodeError
else:
    JSONDecodeError = ValueError


def _isVhdl(path):  # pragma: no cover
    "Uses the file extension to check if the given path is a VHDL file"
    if path.lower().endswith(".vhd"):
        return True
    if path.lower().endswith(".vhdl"):
        return True
    return False


def _isVerilog(path):  # pragma: no cover
    """Uses the file extension to check if the given path is a Verilog
    or SystemVerilog file"""
    if path.lower().endswith(".v"):
        return True
    if path.lower().endswith(".sv"):
        return True
    return False


PARSERS = {
    FileType.vhdl: VhdlParser,
    FileType.verilog: VerilogParser,
    FileType.systemverilog: VerilogParser,
}  # type: Dict[FileType, Type[Union[VhdlParser, VerilogParser]]]


def getSourceParserFromPath(path):  # type: (Path) -> tSourceFile
    """
    Returns either a VhdlParser or VerilogParser based on the path's file
    extension
    """
    return PARSERS[FileType.fromPath(path)](path)


def getSourceFileObjects(kwargs_list, workers=None):
    """
    Gets source file objects by applying each item on kwargs_list as
    kwargs on the source parser class. Uses kwargs['filename'] to
    determine if the source is VHDL or Verilog/SystemVerilog
    """
    pool = Pool(workers)
    async_results = []

    for kwargs in kwargs_list:
        if _isVhdl(kwargs["filename"]):
            cls = VhdlParser
        else:
            cls = VerilogParser
        async_results += [pool.apply_async(cls, kwds=kwargs)]

    pool.close()
    pool.join()
    results = [x.get() for x in async_results]

    return results


def _makeAbsoluteIfNeeded(root, paths):
    # type: (str, Iterable[str]) -> Iterable[str]
    for path in paths:
        if p.isabs(path):
            yield path
        else:
            yield p.join(root, path)


def getIncludedConfigs(search_paths, root_dir="."):
    # type: (Dict[str, Any], str) -> Iterable[Tuple[str, Dict[str, Any]]]
    "Returns configuration contents of included files"
    # Copy the dict to avoid messing up with the caller's env
    #  config = root_config.copy()

    # Will search for inclusion clauses recursivelly but we need to keep
    # track of infinite loops
    checked_paths = set()  # type: Set[str]

    paths = set(_makeAbsoluteIfNeeded(root_dir, search_paths))

    while paths:
        path = paths.pop()
        checked_paths.add(path)

        if not p.exists(path):
            _logger.warning("Skipping included path '%s' (no such file)", path)
            continue

        # Load the config from the file
        try:
            config = json.load(open(path, "r"))
        except JSONDecodeError:
            _logger.warning("Failed to decode file %s", path)
            continue

        extracted_paths = set(
            _makeAbsoluteIfNeeded(p.dirname(path), config.pop("include", ()))
        )
        # Add new paths to the set so they get parsed as well
        paths |= extracted_paths - checked_paths

        yield p.dirname(path), config


class JsonSourceEntry(namedtuple("JsonSourceEntry", ("path", "library", "flags"))):
    """
    Converts different methods of representing a source on the JSON file to a
    saner named tuple
    """

    @classmethod
    def _make(cls, iterable):  # pylint: disable=arguments-differ
        path = iterable
        info = {}

        if not isinstance(path, six.string_types):
            path = iterable[0]
            info = iterable[1]

        library = info.get("library", None)
        flags = info.get("flags", tuple())

        return super(JsonSourceEntry, cls)._make([path, library, flags])


SourceEntry = namedtuple(
    "SourceEntry", ("path", "library", "single_flags", "dependencies_flags")
)


def flattenConfig(root_config, root_path):
    # type: (Dict[str, Any], str) -> Iterable[SourceEntry]
    """
    Expands the given root config and also recursively expands included JSON
    files as well
    """
    # Extract included paths and yield those results first
    include_paths = root_config.pop("include", ())
    for config_path, config in getIncludedConfigs(include_paths, root_path):
        for entry in _expand(config, config_path):
            yield entry

    # Extract sources form the root config last so these are the values that
    # might eventually prevail
    for entry in _expand(root_config, root_path):
        yield entry


def _expand(config, ref_path):
    # type: (Dict[str, Any], str) -> Iterable[SourceEntry]
    """
    Expands the sources defined in the config dict into a list of tuples
    """

    flags = {}

    for filetype in FileType:
        filetype_cfg = config.pop(filetype.value, {}).pop("flags", {})
        flags[filetype] = (
            filetype_cfg.get(BuildFlagScope.single.value, ()),
            filetype_cfg.get(BuildFlagScope.dependencies.value, ()),
            filetype_cfg.get(BuildFlagScope.all.value, ()),
        )

    for entry in config.pop("sources", ()):
        source = JsonSourceEntry._make(entry)
        path = Path(source.path, ref_path)

        filetype = FileType.fromPath(path)

        single_flags = flags[filetype][0]
        dependencies_flags = flags[filetype][1]
        global_flags = flags[filetype][2]

        yield SourceEntry(
            path,
            source.library,
            tuple(global_flags) + tuple(single_flags) + tuple(source.flags),
            tuple(global_flags) + tuple(dependencies_flags),
        )
